#!/usr/bin/env python

import os
import unittest
import matplotlib
matplotlib.use("agg")

from test_data import *
from test_utils import *
from test_directi import *
from test_blast import *

if __name__ == "__main__":
    unittest.main()



r"""
Miscellaneous utility functions and classes
"""

import binascii
import collections
import functools
import json
import operator
import os
import typing
import logging
import re

import igraph
import numpy as np
import scipy.sparse
import sklearn.neighbors
import tqdm

from . import data


log_handler = logging.StreamHandler()
log_handler.setLevel(logging.INFO)
log_handler.setFormatter(logging.Formatter(fmt="[%(levelname)s] %(name)s: %(message)s"))
logger = logging.getLogger("Cell BLAST")
logger.setLevel(logging.INFO)
logger.addHandler(log_handler)


def rand_hex() -> str:
    return binascii.b2a_hex(os.urandom(15)).decode()


class dotdict(dict):
    __getattr__ = dict.get
    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__


def in_ipynb() -> bool:  # pragma: no cover
    r"""
    From a StackOverFlow thread
    """
    try:
        # noinspection PyUnresolvedReferences
        shell = get_ipython().__class__.__name__
        if shell == "ZMQInteractiveShell":
            return True   # Jupyter notebook or qtconsole
        elif shell == "TerminalInteractiveShell":
            return False  # Terminal running IPython
        else:
            return False  # Other type (?)
    except NameError:
        return False      # Probably standard Python interpreter


def smart_tqdm():  # pragma: no cover
    if in_ipynb():
        return tqdm.tqdm_notebook
    return tqdm.tqdm


def with_self_graph(fn: typing.Callable) -> typing.Callable:
    @functools.wraps(fn)
    def wrapped(self, *args, **kwargs):
        with self.graph.as_default():
            return fn(self, *args, **kwargs)
    return wrapped


# Wraps a batch function into minibatch version
def minibatch(batch_size: int, desc: str, use_last: bool = False, progress_bar: bool = True) -> typing.Callable:
    def minibatch_wrapper(func: typing.Callable) -> typing.Callable:
        @functools.wraps(func)
        def wrapped_func(*args, **kwargs):
            total_size = args[0].shape[0]
            if use_last:
                n_batch = np.ceil(
                    total_size / float(batch_size)
                ).astype(np.int)
            else:
                n_batch = max(1, np.floor(
                    total_size / float(batch_size)
                ).astype(np.int))
            for batch_idx in smart_tqdm()(
                    range(n_batch), desc=desc, unit="batches",
                    leave=False, disable=not progress_bar
            ):
                start = batch_idx * batch_size
                end = min((batch_idx + 1) * batch_size, total_size)
                this_args = (item[start:end] for item in args)
                func(*this_args, **kwargs)
        return wrapped_func
    return minibatch_wrapper


# Avoid sklearn warning
def encode_integer(
        label: typing.List[typing.Any], sort: bool = False
) -> typing.Tuple[np.ndarray, np.ndarray]:
    label = np.array(label).ravel()
    classes = np.unique(label)
    if sort:
        classes.sort()
    mapping = {v: i for i, v in enumerate(classes)}
    return np.array([mapping[v] for v in label]), classes


# Avoid sklearn warning
def encode_onehot(
        label: typing.List[typing.Any], sort: bool = False,
        ignore: typing.Optional[typing.Any] = None
) -> scipy.sparse.csr_matrix:
    i, c = encode_integer(label, sort)
    onehot = scipy.sparse.csc_matrix((
        np.ones_like(i, dtype=np.int32), (np.arange(i.size), i)
    ))
    if ignore is None:
        ignore = []
    return onehot[:, ~np.in1d(c, ignore)].tocsr()


class CellTypeDAG(object):

    def __init__(
            self, graph: typing.Optional[igraph.Graph] = None,
            vdict: typing.Optional[typing.Mapping[str, str]] = None
    ) -> None:
        self.graph = igraph.Graph(directed=True) if graph is None else graph
        self.vdict = vdict or {}

    @classmethod
    def load(cls, file: str) -> "CellTypeDAG":
        if file.endswith(".json"):
            return cls.load_json(file)
        elif file.endswith(".obo"):
            return cls.load_obo(file)
        else:
            raise ValueError("Unexpected file format!")

    @classmethod
    def load_json(cls, file: str) -> "CellTypeDAG":
        with open(file, "r") as f:
            d = json.load(f)
        dag = cls()
        dag._build_tree(d)
        return dag

    @classmethod
    def load_obo(cls, file: str) -> "CellTypeDAG":  # Only building on "is_a" relation between CL terms
        import pronto
        ont = pronto.Ontology(file)
        graph, vdict = igraph.Graph(directed=True), {}
        for item in ont:
            if not item.id.startswith("CL"):
                continue
            if "is_obsolete" in item.other and item.other["is_obsolete"][0] == "true":
                continue
            graph.add_vertex(
                name=item.id, cell_ontology_class=item.name,
                desc=str(item.desc), synonyms=[
                    (f"{syn.desc} ({syn.scope})")
                    for syn in item.synonyms
                ]
            )
            assert item.id not in vdict
            vdict[item.id] = item.id
            assert item.name not in vdict
            vdict[item.name] = item.id
            for synonym in item.synonyms:
                if synonym.scope == "EXACT" and synonym.desc != item.name:
                    vdict[synonym.desc] = item.id
        for source in graph.vs:  # pylint: disable=not-an-iterable
            for relation in ont[source["name"]].relations:
                if relation.obo_name != "is_a":
                    continue
                for target in ont[source["name"]].relations[relation]:
                    if not target.id.startswith("CL"):
                        continue
                    graph.add_edge(
                        source["name"],
                        graph.vs.find(name=target.id.split()[0])["name"]  # pylint: disable=no-member
                    )
                    # Split because there are many "{is_infered...}" suffix,
                    # falsely joined to the actual id when pronto parses the
                    # obo file
        return cls(graph, vdict)

    def _build_tree(
            self, d: typing.Mapping[str, str],
            parent: typing.Optional[igraph.Vertex] = None
    ) -> None:  # For json loading
        self.graph.add_vertex(name=d["name"])
        v = self.graph.vs.find(d["name"])
        if parent is not None:
            self.graph.add_edge(v, parent)
        self.vdict[d["name"]] = d["name"]
        if "alias" in d:
            for alias in d["alias"]:
                self.vdict[alias] = d["name"]
        if "children" in d:
            for subd in d["children"]:
                self._build_tree(subd, v)

    def get_vertex(self, name: str) -> igraph.Vertex:
        return self.graph.vs.find(self.vdict[name])

    def is_related(self, name1: str, name2: str) -> bool:
        return self.is_descendant_of(name1, name2) \
            or self.is_ancestor_of(name1, name2)

    def is_descendant_of(self, name1: str, name2: str) -> bool:
        if name1 not in self.vdict or name2 not in self.vdict:
            return False
        shortest_path = self.graph.shortest_paths(
            self.get_vertex(name1), self.get_vertex(name2)
        )[0][0]
        return np.isfinite(shortest_path)

    def is_ancestor_of(self, name1: str, name2: str) -> bool:
        if name1 not in self.vdict or name2 not in self.vdict:
            return False
        shortest_path = self.graph.shortest_paths(
            self.get_vertex(name2), self.get_vertex(name1)
        )[0][0]
        return np.isfinite(shortest_path)

    def conditional_prob(self, name1: str, name2: str) -> float:  # p(name1|name2)
        if name1 not in self.vdict or name2 not in self.vdict:
            return 0
        self.graph.vs["prob"] = 0
        v2_parents = list(self.graph.bfsiter(
            self.get_vertex(name2), mode=igraph.OUT))
        v1_parents = list(self.graph.bfsiter(
            self.get_vertex(name1), mode=igraph.OUT))
        for v in v2_parents:
            v["prob"] = 1
        while True:
            changed = False
            for v1_parent in v1_parents[::-1]:  # Reverse may be more efficient
                if v1_parent["prob"] != 0:
                    continue
                v1_parent["prob"] = np.prod([
                    v["prob"] / v.degree(mode=igraph.IN)
                    for v in v1_parent.neighbors(mode=igraph.OUT)
                ])
                if v1_parent["prob"] != 0:
                    changed = True
            if not changed:
                break
        return self.get_vertex(name1)["prob"]

    def similarity(self, name1: str, name2: str, method: str = "probability") -> float:
        if method == "probability":
            return (
                self.conditional_prob(name1, name2) +
                self.conditional_prob(name2, name1)
            ) / 2
        # if method == "distance":
        #     return self.distance_ratio(name1, name2)
        raise ValueError("Invalid method!")  # pragma: no cover

    def value_reset(self) -> None:
        self.graph.vs["raw_value"] = 0
        self.graph.vs["prop_value"] = 0  # value propagated from children
        self.graph.vs["value"] = 0

    def value_set(self, name: str, value: float) -> None:
        try:
            self.get_vertex(name)["raw_value"] = value
        except KeyError:
            logger.warning("Unknown node name! Doing nothing.")

    def value_update(self) -> None:
        origins = [v for v in self.graph.vs.select(raw_value_gt=0)]
        for origin in origins:
            for v in self.graph.bfsiter(origin, mode=igraph.OUT):
                if v != origin:  # bfsiter includes the vertex self
                    v["prop_value"] += origin["raw_value"]
        self.graph.vs["value"] = list(map(
            operator.add, self.graph.vs["raw_value"],
            self.graph.vs["prop_value"]
        ))

    def best_leaves(
            self, thresh: float = 0.5, min_path: int = 4,
            retrieve: str = "name"
    ) -> typing.List[str]:
        subgraph = self.graph.subgraph(self.graph.vs.select(value_gt=thresh))
        leaves, max_value = [], 0
        for leaf in subgraph.vs.select(lambda v: v.indegree() == 0):
            if self.longest_paths_to_root(leaf["name"]) < min_path:
                continue
            if leaf["value"] > max_value:
                max_value = leaf["value"]
                leaves = [leaf[retrieve]]
            elif leaf["value"] == max_value:
                leaves.append(leaf[retrieve])
        return leaves

    def cal_longest_paths_to_root(self, weight: float = 1.0) -> None:
        self.graph.vs["longest_paths_to_root"] = -np.inf
        roots = self.graph.vs.select(lambda v: v.outdegree() == 0)
        for root in roots:
            root["longest_paths_to_root"] = 0
        self.graph.es["weight"] = weight
        for vertex in self.graph.vs[self.graph.topological_sorting(mode=igraph.IN)]:
            for neighbor in self.graph.vs[self.graph.neighborhood(vertex, mode=igraph.IN)]:
                if neighbor == vertex:
                    continue
                if neighbor["longest_paths_to_root"] < vertex["longest_paths_to_root"] + \
                        self.graph[neighbor, vertex]:
                    neighbor["longest_paths_to_root"] = vertex["longest_paths_to_root"] + \
                        self.graph[neighbor, vertex]

    def longest_paths_to_root(self, name: str) -> int:
        if "longest_paths_to_root" not in self.get_vertex(name).attribute_names():
            self.cal_longest_paths_to_root()
        return self.get_vertex(name)["longest_paths_to_root"]


class DataDict(collections.OrderedDict):

    def shuffle(self, random_state: np.random.RandomState = np.random) -> "DataDict":
        shuffled = DataDict()
        shuffle_idx = None
        for item in self:
            shuffle_idx = random_state.permutation(self[item].shape[0]) \
                if shuffle_idx is None else shuffle_idx
            shuffled[item] = self[item][shuffle_idx]
        return shuffled

    @property
    def size(self) -> int:
        data_size = set([item.shape[0] for item in self.values()])
        if data_size:
            assert len(data_size) == 1
            return data_size.pop()
        return 0

    @property
    def shape(self) -> typing.List[int]:  # Compatibility with numpy arrays
        return [self.size]

    def __getitem__(
            self, fetch: typing.Union[str, slice, np.ndarray]
    ) -> typing.Union["DataDict", np.ndarray]:
        if isinstance(fetch, (slice, np.ndarray)):
            return DataDict([
                (item, self[item][fetch]) for item in self
            ])
        return super(DataDict, self).__getitem__(fetch)


def densify(arr: typing.Union[np.ndarray, scipy.sparse.spmatrix]) -> np.ndarray:
    if scipy.sparse.issparse(arr):
        return arr.toarray()
    return arr


def empty_safe(fn: typing.Callable, dtype: type):
    def _fn(x):
        if x.size:
            return fn(x)
        return x.astype(dtype)
    return _fn


def subsample_molecules(
        ds: data.ExprDataSet, ratio: float = 0.9, random_seed: int = 0
) -> data.ExprDataSet:
    r"""
    Subsample molecules

    Parameters
    ----------
    ds
        Dataset to be subsampled.
    ratio
        Subsample ratio.
    random_seed
        Random seed.

    Returns
    -------
    subsampled
        Subsampled dataset
    """
    random_state = np.random.RandomState(seed=random_seed)
    ds = ds.copy()  # Shallow
    x = ds.exprs.copy()  # Deep
    if not np.issubdtype(x.dtype.type, np.integer):
        x_int = np.round(x).astype(int)
        if np.abs(x - x_int).max() > 1e-8:
            logger.warning("Input not integer! Rounding to nearest integer.")
        x = x_int
    if scipy.sparse.issparse(x):
        x.data = random_state.binomial(x.data, ratio)
        x.eliminate_zeros()
    ds.exprs = x
    return ds


def split_molecules(
        ds: data.ExprDataSet, val_split: float = 0.1, random_seed: int = 0
) -> typing.Tuple[data.ExprDataSet, data.ExprDataSet]:
    r"""
    Molecular split (only disjoint split, i.e. no overlap between splits).

    Parameters
    ----------
    ds
        Dataset to be splitted.
    val_split
        Ratio of validation set.
    random_seed
        Random seed.

    Returns
    -------
    train
        Training dataset
    val
        Validation dataset
    """
    train_ds = subsample_molecules(ds, 1 - val_split, random_seed=random_seed)
    val_ds = ds.copy()
    val_ds.exprs = np.round(ds.exprs).astype(int) - train_ds.exprs
    if scipy.sparse.issparse(val_ds.exprs):
        val_ds.exprs.eliminate_zeros()
    return train_ds, val_ds


def neighbor_stability(
        ds: data.ExprDataSet,
        metric: str = "minkowski", k: typing.Union[int, float] = 0.01,
        used_genes: typing.Optional[typing.List[str]] = None, n_jobs: int = 1,
        subsample_ratio: float = 0.8, n_repeats: int = 5, random_seed: int = 0
) -> scipy.sparse.csr_matrix:
    r"""
    Get original space nearest neighbor stability across molecular subsampling.

    Parameters
    ----------
    ds
        Dataset being considered
    k
        Number (if k is an integer greater than 1) or fraction in total data
        (if k is a float less than 1) of nearest neighbors to consider.
    metric
        Distance metric to be used.
        See :class:`sklearn.neighbors.NearestNeighbors` for available options.
    used_genes
        A subset of genes to be used when computing distance.
    n_jobs
        Number of parallel jobs to use when doing nearest neighbor search.
        See :class:`sklearn.neighbors.NearestNeighbors` for details.
    subsample_ratio
        Subsample ratio.
    n_repeats
        Number of subsample repeats.
    random_seed
        Random seed.

    Returns
    -------
    nng
        Stable nearest neighbor graph
    """
    n = ds.shape[0]
    k = n * k if k < 1 else k
    k = np.round(k).astype(np.int)
    nng = scipy.sparse.csr_matrix((n, n))
    for i in range(n_repeats):
        subsample = subsample_molecules(
            ds, ratio=subsample_ratio, random_seed=random_seed + i
        )
        subsample = subsample.normalize()
        subsample.exprs = np.log1p(subsample.exprs)
        if used_genes is not None:
            subsample = subsample[:, used_genes]
        subsample = subsample.exprs.toarray()
        nn = sklearn.neighbors.NearestNeighbors(
            n_neighbors=min(n, k + 1), metric=metric, n_jobs=n_jobs
        ).fit(subsample)
        nng += nn.kneighbors_graph(subsample) - scipy.sparse.eye(n)
    return nng / n_repeats


def scope_free(x: str) -> str:
    return re.sub(r"[^A-Za-z0-9_.\\-]", "_", x)


def isnan(x: typing.Any) -> bool:
    try:
        return np.isnan(x)
    except Exception:
        return False


isnan = empty_safe(np.vectorize(isnan), bool)
decode = empty_safe(np.vectorize(lambda _x: _x.decode("utf-8")), str)
encode = empty_safe(np.vectorize(lambda _x: str(_x).encode("utf-8")), "S")
upper = empty_safe(np.vectorize(lambda x: str(x).upper()), str)
lower = empty_safe(np.vectorize(lambda x: str(x).lower()), str)
tostr = empty_safe(np.vectorize(str), str)



r"""
Global configuration
"""

RANDOM_SEED = 0
N_JOBS = 1

_USE_GLOBAL = "__UsE_gLoBaL__"
_NAN_REPLACEMENT = "__nAn_RePlAcEmEnT__"

H5_COMPRESS_OPTS = {
    "compression": "gzip",
    "compression_opts": 7,
}
H5_TRACK_OPTS = {
    "track_times": False
}



r"""
The Cell_BLAST package
"""

from .utils import in_ipynb

if not in_ipynb():
    import matplotlib
    matplotlib.use("agg")

from . import (blast, config, data, directi, latent, metrics, prob, rmbatch,
               utils)


name = "Cell_BLAST"

__copyright__ = "2020, Gao Lab"

__author__ = "Zhijie Cao"

__all__ = [
    "blast",
    "data",
    "directi",
    "latent",
    "metrics",
    "prob",
    "rmbatch",
    "utils",
    "config"
]

__version__ = "0.3.6"



import os
import sys
import unittest

if os.environ.get("TEST_MODE", "INSTALL") == "DEV":
    sys.path.insert(0, "..")
import Cell_BLAST as cb


class TestJsonCellTypeTree(unittest.TestCase):

    def setUp(self):
        self.dag = cb.utils.CellTypeDAG.load("./tree.json")

    def test_load(self):
        with self.assertRaises(ValueError):
            dag = cb.utils.CellTypeDAG.load("./asd.asd")

    def test_alias(self):
        self.assertIn("root", self.dag.vdict)
        l1 = ["a", "b", "c", "d", "e", "f", "g", "h"]
        l2 = ["A", "B", "C", "D", "E", "F", "G", "H"]
        for n1, n2 in zip(l1, l2):
            self.assertEqual(self.dag.get_vertex(n1), self.dag.get_vertex(n2))

    def test_relation(self):
        self.assertFalse(self.dag.is_related("b", "c"))
        self.assertFalse(self.dag.is_related("d", "i"))
        self.assertTrue(self.dag.is_related("g", "e"))
        self.assertTrue(self.dag.is_descendant_of("g", "e"))
        self.assertFalse(self.dag.is_descendant_of("e", "g"))
        self.assertTrue(self.dag.is_ancestor_of("e", "g"))
        self.assertFalse(self.dag.is_ancestor_of("g", "e"))

    def test_similarity(self):
        self.assertAlmostEqual(self.dag.similarity("g", "h"), 0.5)
        self.assertAlmostEqual(self.dag.similarity("g", "f"), 0.75)
        self.assertAlmostEqual(self.dag.similarity("g", "e"), 0.75)
        self.assertAlmostEqual(self.dag.similarity("b", "d"), 0.25)
        self.assertAlmostEqual(self.dag.similarity("h", "c"), 1 / 6)
        self.assertAlmostEqual(self.dag.similarity("h", "i"), 0)

    def test_value(self):
        self.dag.value_reset()
        for item in self.dag.vdict:
            self.assertEqual(self.dag.get_vertex(item)["value"], 0)
        self.dag.value_set("g", 5)
        self.dag.value_set("f", 2)
        self.dag.value_set("c", 3)
        self.dag.value_update()
        self.assertEqual(self.dag.get_vertex("root")["value"], 10)
        self.assertEqual(self.dag.get_vertex("a")["value"], 3)
        self.assertEqual(self.dag.get_vertex("b")["value"], 0)
        self.assertEqual(self.dag.get_vertex("c")["value"], 3)
        self.assertEqual(self.dag.get_vertex("d")["value"], 0)
        self.assertEqual(self.dag.get_vertex("e")["value"], 7)
        self.assertEqual(self.dag.get_vertex("f")["value"], 7)
        self.assertEqual(self.dag.get_vertex("g")["value"], 5)
        self.assertEqual(self.dag.get_vertex("h")["value"], 0)
        result = self.dag.best_leaves(4, min_path=0)
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0], "g")
        result = self.dag.best_leaves(6, min_path=0)
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0], "f")


class TestOBOCellTypeTree(TestJsonCellTypeTree):

    def setUp(self):
        self.dag = cb.utils.CellTypeDAG.load("./tree.obo")

    def test_value(self):
        self.dag.value_reset()
        for item in self.dag.vdict:
            self.assertEqual(self.dag.get_vertex(item)["value"], 0)
        self.dag.value_set("g", 5)
        self.dag.value_set("f", 2)
        self.dag.value_set("c", 3)
        self.dag.value_update()
        self.assertEqual(self.dag.get_vertex("root")["value"], 10)
        self.assertEqual(self.dag.get_vertex("a")["value"], 3)
        self.assertEqual(self.dag.get_vertex("b")["value"], 0)
        self.assertEqual(self.dag.get_vertex("c")["value"], 3)
        self.assertEqual(self.dag.get_vertex("d")["value"], 0)
        self.assertEqual(self.dag.get_vertex("e")["value"], 7)
        self.assertEqual(self.dag.get_vertex("f")["value"], 7)
        self.assertEqual(self.dag.get_vertex("g")["value"], 5)
        self.assertEqual(self.dag.get_vertex("h")["value"], 0)
        result = self.dag.best_leaves(4, retrieve="cell_ontology_class", min_path=0)
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0], "g")
        result = self.dag.best_leaves(6, retrieve="cell_ontology_class", min_path=0)
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0], "f")


class TestOBOCellTypeDAG(TestJsonCellTypeTree):

    def setUp(self):
        self.dag = cb.utils.CellTypeDAG.load("./dag.obo")

    def test_similarity(self):
        self.assertAlmostEqual(self.dag.similarity("g", "h"), 7 / 24)
        self.assertAlmostEqual(self.dag.similarity("g", "f"), 0.75)
        self.assertAlmostEqual(self.dag.similarity("g", "e"), 0.75)
        self.assertAlmostEqual(self.dag.similarity("b", "d"), 0.25)
        self.assertAlmostEqual(self.dag.similarity("h", "c"), 1 / 12)
        self.assertAlmostEqual(self.dag.similarity("h", "i"), 0)

    def test_value(self):
        self.dag.value_reset()
        for item in self.dag.vdict:
            self.assertEqual(self.dag.get_vertex(item)["value"], 0)
        self.dag.value_set("g", 5)
        self.dag.value_set("f", 2)
        self.dag.value_set("c", 3)
        self.dag.value_update()
        self.assertEqual(self.dag.get_vertex("root")["value"], 10)
        self.assertEqual(self.dag.get_vertex("a")["value"], 3)
        self.assertEqual(self.dag.get_vertex("b")["value"], 0)
        self.assertEqual(self.dag.get_vertex("c")["value"], 3)
        self.assertEqual(self.dag.get_vertex("d")["value"], 3)
        self.assertEqual(self.dag.get_vertex("e")["value"], 7)
        self.assertEqual(self.dag.get_vertex("f")["value"], 7)
        self.assertEqual(self.dag.get_vertex("g")["value"], 5)
        self.assertEqual(self.dag.get_vertex("h")["value"], 0)
        result = self.dag.best_leaves(4, retrieve="cell_ontology_class", min_path=0)
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0], "g")
        result = self.dag.best_leaves(6, retrieve="cell_ontology_class", min_path=0)
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0], "f")


# class TestOBOCellOntology(unittest.TestCase):
#
#     def setUp(self):
#         self.dag = utils.CellTypeDAG.load(
#             "../Datasets/cell_ontology/github/cl.obo")
#
#     def test_pass(self):
#         pass

# # Snippet for drawing subgraph related to vertex v
# igraph.drawing.plot(dag.graph.subgraph(
#     [item.index for item in dag.graph.bfsiter(v.index, mode=igraph.OUT)] +
#     [item.index for item in dag.graph.bfsiter(v.index, mode=igraph.IN)],
# ), "test.svg", margin=(100, 100, 100, 100))


if __name__ == "__main__":
    unittest.main()


r"""
Functions for computing benchmark metrics
"""

import typing

import numpy as np
import pandas as pd
import scipy.sparse
import sklearn.metrics
import sklearn.neighbors
import igraph

from . import blast, utils

_identity = lambda x, y: 1 if x == y else 0


#===============================================================================
#
#  Cluster based metrics
#
#===============================================================================
def confusion_matrix(x: np.ndarray, y: np.ndarray) -> pd.DataFrame:
    r"""
    Reimplemented this because sklearn.metrics.confusion_matrix
    does not provide row names and column names.
    """
    x, x_c = utils.encode_integer(x)
    y, y_c = utils.encode_integer(y)
    unique_x, unique_y = np.unique(x), np.unique(y)
    cm = np.empty((len(unique_x), len(unique_y)), dtype=np.int)
    for i in unique_x:
        for j in unique_y:
            cm[i, j] = np.sum((x == i) & (y == j))
    return pd.DataFrame(data=cm, index=x_c, columns=y_c)


def class_specific_accuracy(
        true: np.ndarray, pred: np.ndarray, expectation: pd.DataFrame
) -> pd.DataFrame:
    df = pd.DataFrame(index=np.unique(true), columns=["number", "accuracy"])
    expectation = expectation.astype(np.bool)
    for c in df.index:
        true_mask = true == c
        pred_mask = np.in1d(pred, expectation.columns[expectation.loc[c]])
        df.loc[c, "number"] = true_mask.sum()
        df.loc[c, "accuracy"] = np.logical_and(pred_mask, true_mask).sum() / df.loc[c, "number"]
    return df


def mean_balanced_accuracy(
        true: np.ndarray, pred: np.ndarray, expectation: pd.DataFrame,
        population_weighed: bool = False
) -> float:
    df = class_specific_accuracy(true, pred, expectation)
    if population_weighed:
        return (df["accuracy"] * df["number"]).sum() / df["number"].sum()
    return df["accuracy"].mean()


def cl_accuracy(
        cl_dag: utils.CellTypeDAG,
        source: np.ndarray, target: np.ndarray,
        ref_cl_list: typing.List[str]  # a list of unique cl in ref
) -> pd.DataFrame:
    if len(source) != len(target):
        raise ValueError("Invalid input: different cell number.")
    positive_cl = []
    negative_cl = []
    for query_cl in np.unique(source):
        for ref_cl in ref_cl_list:
            if cl_dag.is_related(query_cl, ref_cl):
                positive_cl.append(query_cl)
                break
        else:
            negative_cl.append(query_cl)

    # sensitivity
    accuracy_dict = {}
    ref_cl_set = {cl_dag.get_vertex(item) for item in ref_cl_list}
    for query_cl in positive_cl:
        target_cl = target[source == query_cl]
        cl_accuracy = np.zeros(target_cl.size)
        cache = {}  # avoid repeated computation
        for i, _target_cl in enumerate(target_cl):
            if _target_cl in cache:
                cl_accuracy[i] = cache[_target_cl]
                continue
            if cl_dag.is_descendant_of(_target_cl, query_cl):
                cache[_target_cl] = 1  # equal or descendant results as 1
            elif cl_dag.is_ancestor_of(_target_cl, query_cl):
                intermediates = set.intersection(
                    set(cl_dag.graph.bfsiter(cl_dag.get_vertex(query_cl), mode=igraph.OUT)),
                    set(cl_dag.graph.bfsiter(cl_dag.get_vertex(_target_cl), mode=igraph.IN))
                )
                cache[_target_cl] = np.mean([
                    len(list(cl_dag.graph.bfsiter(intermediate, mode=igraph.IN))) /
                    len(list(cl_dag.graph.bfsiter(cl_dag.get_vertex(_target_cl), mode=igraph.IN)))
                    for intermediate in intermediates
                    if ref_cl_set.intersection(set(cl_dag.graph.bfsiter(intermediate, mode=igraph.IN)))
                ])
            else:
                cache[_target_cl] = 0
            cl_accuracy[i] = cache[_target_cl]
        accuracy_dict[query_cl] = [cl_accuracy.size, cl_accuracy.mean(), True]

    # specificity
    for query_cl in negative_cl:
        target_cl = target[source == query_cl]
        cl_accuracy = np.in1d(target_cl, ["rejected", "unassigned"])
        accuracy_dict[query_cl] = [cl_accuracy.size, cl_accuracy.mean(), False]

    return pd.DataFrame(accuracy_dict, index=("cell number", "accuracy", "positive")).T


def cl_mba(
        cl_dag: utils.CellTypeDAG,
        source: np.ndarray, target: np.ndarray,
        ref_cl_list: typing.List[str]
) -> float:
    accuracy_df = cl_accuracy(cl_dag=cl_dag, source=source, target=target, ref_cl_list=ref_cl_list)
    return accuracy_df["accuracy"].mean()


#===============================================================================
#
#  Distance based metrics
#
#===============================================================================
def nearest_neighbor_accuracy(
        x: np.ndarray, y: np.ndarray, metric: str = "minkowski",
        similarity: typing.Callable = _identity, n_jobs: int = 1
) -> float:
    nearestNeighbors = sklearn.neighbors.NearestNeighbors(
        n_neighbors=2, metric=metric, n_jobs=n_jobs)
    nearestNeighbors.fit(x)
    nni = nearestNeighbors.kneighbors(x, return_distance=False)
    return np.vectorize(similarity)(y, y[nni[:, 1].ravel()]).mean()


def mean_average_precision_from_latent(
        x: np.ndarray, y: np.ndarray, p: typing.Optional[np.ndarray] = None,
        k: float = 0.01, metric: str = "minkowski", posterior_metric: str = "npd_v1",
        similarity: typing.Callable = _identity, n_jobs: int = 1
) -> float:
    if k < 1:
        k = y.shape[0] * k
    k = np.round(k).astype(np.int)
    nearestNeighbors = sklearn.neighbors.NearestNeighbors(
        n_neighbors=min(y.shape[0], k + 1), metric=metric, n_jobs=n_jobs)
    nearestNeighbors.fit(x)
    nni = nearestNeighbors.kneighbors(x, return_distance=False)
    if p is not None:
        posterior_metric = getattr(blast, posterior_metric)
        pnnd = np.empty_like(nni, np.float32)
        for i in range(pnnd.shape[0]):
            for j in range(pnnd.shape[1]):
                pnnd[i, j] = posterior_metric(
                    x[i], x[nni[i, j]],
                    p[i], p[nni[i, j]]
                )
            nni[i] = nni[i][np.argsort(pnnd[i])]
    return mean_average_precision(y, y[nni[:, 1:]], similarity=similarity)


def average_silhouette_score(x: np.ndarray, y: np.ndarray) -> float:
    return sklearn.metrics.silhouette_score(x, y)


def seurat_alignment_score(
        x: np.ndarray, y: np.ndarray, k: float = 0.01, n: int = 1,
        metric: str = "minkowski", random_seed: typing.Optional[int] = None,
        n_jobs: int = 1
) -> float:
    random_state = np.random.RandomState(random_seed)
    idx_list = [np.where(y == _y)[0] for _y in np.unique(y)]
    subsample_size = min(idx.size for idx in idx_list)
    subsample_scores = []
    for _ in range(n):
        subsample_idx_list = [
            random_state.choice(idx, subsample_size, replace=False)
            for idx in idx_list
        ]
        subsample_y = y[np.concatenate(subsample_idx_list)]
        subsample_x = x[np.concatenate(subsample_idx_list)]
        _k = subsample_y.shape[0] * k if k < 1 else k
        _k = np.round(_k).astype(np.int)
        nearestNeighbors = sklearn.neighbors.NearestNeighbors(
            n_neighbors=min(subsample_y.shape[0], _k + 1),
            metric=metric, n_jobs=n_jobs
        )
        nearestNeighbors.fit(subsample_x)
        nni = nearestNeighbors.kneighbors(subsample_x, return_distance=False)
        same_y_hits = (
            subsample_y[nni[:, 1:]] == np.expand_dims(subsample_y, axis=1)
        ).sum(axis=1).mean()
        subsample_scores.append(
            (_k - same_y_hits) * len(idx_list) /
            (_k * (len(idx_list) - 1))
        )
    return np.mean(subsample_scores)


def batch_mixing_entropy(
        x: np.ndarray, y: np.ndarray, boots: int = 100,
        sample_size: int = 100, k: int = 100, metric: str = "minkowski",
        random_seed: typing.Optional[int] = None, n_jobs: int = 1
) -> float:
    random_state = np.random.RandomState(random_seed)
    batches = np.unique(y)
    entropy = 0
    for _ in range(boots):
        bootsamples = random_state.choice(
            np.arange(x.shape[0]), sample_size, replace=False)
        subsample_x = x[bootsamples]
        neighbor = sklearn.neighbors.NearestNeighbors(
            n_neighbors=k, metric=metric, n_jobs=n_jobs
        )
        neighbor.fit(x)
        nn = neighbor.kneighbors(subsample_x, return_distance=False)
        for i in range(sample_size):
            for batch in batches:
                b = len(np.where(y[nn[i, :]] == batch)[0]) / k
                if b == 0:
                    entropy = entropy
                else:
                    entropy = entropy + b * np.log(b)
    entropy = -entropy / (boots * sample_size)
    return entropy


#===============================================================================
#
#  Ranking based metrics
#
#===============================================================================
def _average_precision(r: np.ndarray) -> float:
    cummean = np.cumsum(r) / (np.arange(r.size) + 1)
    mask = r > 0
    if np.any(mask):
        return cummean[mask].mean()
    return 0.0


def mean_average_precision(
        true: np.ndarray, hits: np.ndarray,
        similarity: typing.Callable[[typing.Any, typing.Any], float] = _identity
) -> float:
    r"""
    Mean average precision

    Parameters
    ----------
    true
        True label
    hits
        Hit labels
    similarity
        Function that defines the similarity between labels

    Returns
    -------
    map
        Mean average precision
    """
    r = np.apply_along_axis(np.vectorize(similarity), 0, hits, true)
    return np.apply_along_axis(_average_precision, 1, r).mean()


#===============================================================================
#
#  Structure preservation
#
#===============================================================================
def avg_neighbor_jacard(
        x: np.ndarray, y: np.ndarray,
        x_metric: str = "minkowski", y_metric: str = "minkowski",
        k: typing.Union[int, float] = 0.01, n_jobs: int = 1
) -> float:
    r"""
    Average neighborhood Jacard index.

    Parameters
    ----------
    x
        First feature space.
    y
        Second feature space.
    x_metric
        Distance metric to use in first feature space.
        See :class:`sklearn.neighbors.NearestNeighbors` for available options.
    y_metric
        Distance metric to use in second feature space.
        See :class:`sklearn.neighbors.NearestNeighbors` for available options.
    k
        Number (if k is an integer greater than 1) or fraction in total data
        (if k is a float less than 1) of nearest neighbors to consider.
    n_jobs
        Number of parallel jobs to use when doing nearest neighbor search.
        See :class:`sklearn.neighbors.NearestNeighbors` for details.

    Returns
    -------
    jacard
        Average neighborhood Jacard index.
    """
    if x.shape[0] != y.shape[0]:
        raise ValueError("Inconsistent shape!")
    n = x.shape[0]
    k = n * k if k < 1 else k
    k = np.round(k).astype(np.int)
    nn = sklearn.neighbors.NearestNeighbors(
        n_neighbors=min(n, k + 1), metric=x_metric, n_jobs=n_jobs
    ).fit(x)
    nni_x = nn.kneighbors(x, return_distance=False)[:, 1:]
    nn = sklearn.neighbors.NearestNeighbors(
        n_neighbors=min(n, k + 1), metric=y_metric, n_jobs=n_jobs
    ).fit(y)
    nni_y = nn.kneighbors(y, return_distance=False)[:, 1:]
    jacard = np.array([
        np.intersect1d(i, j).size / np.union1d(i, j).size
        for i, j in zip(nni_x, nni_y)
    ])
    return jacard.mean()


def jacard_index(
        x: scipy.sparse.csr_matrix, y: scipy.sparse.csr_matrix
) -> np.ndarray:
    r"""
    Compute Jacard index between two nearest neighbor graphs

    Parameters
    ----------
    x
        First nearest neighbor graph
    y
        Second nearest neighbor graph

    Returns
    -------
    jacard
        Jacard index for each row
    """
    xy = x + y
    return np.asarray((xy == 2).sum(axis=1)).ravel() / \
        np.asarray((xy > 0).sum(axis=1)).ravel()


def neighbor_preservation_score(
        x: np.ndarray, nng: scipy.sparse.spmatrix,
        metric: str = "minkowski", k: typing.Union[int, float] = 0.01,
        n_jobs: int = 1
) -> float:
    if not x.shape[0] == nng.shape[0] == nng.shape[1]:
        raise ValueError("Inconsistent shape!")
    n = x.shape[0]
    k = n * k if k < 1 else k
    k = np.round(k).astype(np.int)
    nn = sklearn.neighbors.NearestNeighbors(
        n_neighbors=min(n, k + 1), metric=metric, n_jobs=n_jobs
    ).fit(x)
    nni = nn.kneighbors(x, return_distance=False)[:, 1:]
    ap = np.array([
        _average_precision(_nng.toarray().ravel()[_nni])
        for _nng, _nni in zip(nng, nni)
    ])
    max_ap = np.array([
        _average_precision(np.sort(
            _nng.toarray().ravel()
        )[::-1][:nni.shape[1]]) for _nng in nng
    ])
    ap /= max_ap
    return ap.mean()



r"""
Dataset utilities
"""

import collections
import concurrent.futures
import copy
import functools
import os
import typing

import anndata
import h5py
import loompy
import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import scipy.sparse
import scipy.stats
import seaborn as sns
import sklearn.metrics

from . import config, utils


class ExprDataSet(object):

    r"""
    Main data class, which is based on the data structure of
    :class:`anndata.AnnData`.
    Note that the data is always assumed to be scRNA-seq, so the stored
    matrix is always the expression matrix. The ``obs`` slot stores meta
    information of cells, and the ``var`` slot stores meta information of genes.
    The ``uns`` slot stores other unstructure data, e.g. list of most
    informative genes, etc.

    Parameters
    ----------
    exprs
        An :math:`obs \times var` expression matrix in the form of
        either a numpy array or a scipy sparse matrix.
    obs
        Cell meta table. Each row corresponds to a row in ``exprs``.
    var
        Gene meta table. Each row corresponds to a column in ``exprs``.
    uns
        Unstructured meta information, e.g. list of most informative genes.

    Examples
    --------

    An :class:`ExprDataSet` object can be constructed from an expression matrix,
    an observation (cell) meta table, a variable (gene) meta table, and some
    unstructured data:

    >>> data_obj = Cell_BLAST.data.ExprDataSet(exprs, obs, var, uns)

    Or, if you have an :class:`anndata.AnnData` object or a
    :class:`loompy.loompy.LoomConnection` object (to a loom file), you can
    directly convert them to an :class:`ExprDataSet` object using the following
    methods:

    >>> data_obj = Cell_BLAST.data.ExprDataSet.from_anndata(anndata_obj)
    >>> data_obj = Cell_BLAST.data.ExprDataSet.from_loom(loomconnection_obj)

    It's also possible to convert in the opposite direction:

    >>> anndata_obj = data_obj.to_anndata()
    >>> loomconnection = data_obj.to_loom(filename)

    :class:`ExprDataSet` objects support many forms of slicing, including
    numeric range, numeric index, boolean mask, and obs/var name selection:

    >>> subdata_obj = data_obj[0:10, np.arange(10)]
    >>> subdata_obj = data_obj[
    ...     data_obj.obs["cell_ontology_class"] == "endothelial",
    ...     ["gene_1", "gene_2"]
    ... ]

    Note that in variable name selection, if a variable does not exist in the
    original dataset, it will be filled with zeros in the returned dataset,
    with a warning message.

    :class:`ExprDataSet` objects also support saving and loading:

    >>> data_obj.write_dataset("data.h5")
    >>> data_obj = Cell_BLAST.data.ExprDataSet.read_dataset("data.h5")

    Some other utilities used in the Cell_BLAST pipeline are also supported,
    including but not limited to:

    Dataset merging

    >>> combined_data_obj = Cell_BLAST.data.ExprDataSet.merge_datasets({
    ...     "data1": data_obj1,
    ...     "data2": data_obj2,
    ...     "data3": data_obj3
    ... })

    Data visualization

    >>> data_obj.latent = latent_matrix
    >>> ax = data_obj.visualize_latent("cell_type")
    >>> ax = data_obj.violin("cell_type", "gene_name")
    >>> ax = data_obj.obs_correlation_heatmap()

    Find markers:

    >>> marker_dict = data_obj.fast_markers("cell_type")
    """

    def __init__(
            self, exprs: typing.Union[np.ndarray, scipy.sparse.spmatrix],
            obs: pd.DataFrame, var: pd.DataFrame, uns: typing.Mapping
    ) -> None:
        assert exprs.shape[0] == obs.shape[0] and exprs.shape[1] == var.shape[0]
        if scipy.sparse.issparse(exprs):
            self.exprs = exprs.tocsr()
        else:
            self.exprs = exprs
        self.obs = obs
        self.var = var
        self.uns = uns

    @property
    def X(self) -> typing.Union[np.ndarray, scipy.sparse.spmatrix]:  # For compatibility with `AnnData`
        r"""
        :math:`obs \times var` expression matrix, same as ``exprs``
        """
        return self.exprs

    @property
    def obs_names(self) -> pd.Index:
        r"""
        Name of observations (cells)
        """
        return self.obs.index

    @obs_names.setter
    def obs_names(self, new_names: np.ndarray) -> None:
        assert len(new_names) == self.obs.shape[0]
        self.obs.index = new_names

    @property
    def var_names(self) -> pd.Index:
        r"""
        Name of variables (genes)
        """
        return self.var.index

    @var_names.setter
    def var_names(self, new_names: np.ndarray) -> None:
        assert len(new_names) == self.var.shape[0]
        self.var.index = new_names

    @property
    def shape(self) -> typing.Tuple[int, int]:
        r"""
        Shape of dataset (:math:`obs \times var`)
        """
        return self.exprs.shape

    @property
    def latent(self) -> np.ndarray:
        r"""
        Latent space coordinate. Must have the same number of observations
        (cells) as the expression data.
        """
        mask = np.vectorize(lambda x: x.startswith("latent_"))(self.obs.columns)
        if np.any(mask):
            return self.obs.loc[:, np.vectorize(lambda x: f"latent_{x}")(
                np.arange(mask.sum()) + 1
            )].values
        else:
            raise ValueError("No latent has been registered!")

    @latent.setter
    def latent(self, latent: np.ndarray) -> None:
        for col in self.obs.columns:  # Remove previous result
            if col.startswith("latent_"):
                del self.obs[col]
            if col.startswith("tSNE"):
                del self.obs[col]
            if col.startswith("UMAP"):
                del self.obs[col]
        assert latent.shape[0] == self.shape[0]
        columns = np.vectorize(
            lambda x: f"latent_{x}"
        )(np.arange(latent.shape[1]) + 1)
        latent_df = pd.DataFrame(latent, index=self.obs_names, columns=columns)
        self.obs = pd.concat([self.obs, latent_df], axis=1)

    def normalize(self, target: float = 10000.0) -> "ExprDataSet":
        r"""
        Obs-wise (cell-wise) normalization if the expression matrix.
        Note that only the matrix gets copied in the returned dataset, but meta
        tables are not (only references to meta tables in the original dataset).

        Parameters
        ----------
        target
            Target value of normalization.

        Returns
        -------
        normalized
            Normalized ExprDataSet object.
        """
        import sklearn.preprocessing
        tmp = self.copy()
        if "__libsize__" not in self.obs.columns:
            tmp.exprs = sklearn.preprocessing.normalize(
                tmp.exprs, norm="l1", copy=True
            ) * target
        else:
            normalizer = target / np.expand_dims(
                tmp.obs["__libsize__"].to_numpy(), axis=1)
            tmp.exprs = tmp.exprs.multiply(normalizer) \
                if scipy.sparse.issparse(tmp.exprs) \
                else tmp.exprs * normalizer
        return tmp

    def __getitem__(self, slices) -> "ExprDataSet":
        r"""
        2-d slicing by numeric range, numeric index, boolean mask and obs/var names.
        """
        obs_slice, var_slice = slices

        # var splicing
        simple_flag = False
        if isinstance(var_slice, slice):
            simple_flag = True
        else:  # iterable
            var_slice = np.array(var_slice)
            if var_slice.ndim == 0:
                var_slice = np.expand_dims(var_slice, axis=0)
            if var_slice.size == 0:
                var_slice = var_slice.astype(int)
                simple_flag = True
            if np.issubdtype(var_slice.dtype.type, np.bool_):
                var_slice = np.where(var_slice)[0]
            if np.issubdtype(var_slice.dtype.type, np.integer):
                simple_flag = True

        if simple_flag:
            exprs = self.exprs.tocsc()[:, var_slice].tocsr() \
                if scipy.sparse.issparse(self.exprs) else self.exprs[:, var_slice]
            var = self.var.iloc[var_slice, :]
        else:  # Name slicing
            new_var_names = np.setdiff1d(var_slice, self.var_names)
            all_var_names = np.concatenate([
                self.var_names.values, new_var_names])
            if new_var_names.size > 0:  # pragma: no cover
                utils.logger.warning(
                    "%d out of %d variables are not found, will be set to zero!",
                    len(new_var_names), len(var_slice)
                )
                utils.logger.info(str(new_var_names.tolist()).strip("[]"))
            idx = np.vectorize(
                lambda x: np.where(all_var_names == x)[0][0]
            )(var_slice)
            exprs = scipy.sparse.hstack([
                self.exprs.tocsc(),
                scipy.sparse.csc_matrix((
                    self.exprs.shape[0], len(new_var_names)
                ), dtype=self.exprs.dtype)
            ])[:, idx].tocsr() if scipy.sparse.issparse(self.exprs) else np.concatenate([
                self.exprs, np.zeros((
                    self.exprs.shape[0], len(new_var_names)
                ), dtype=self.exprs.dtype)
            ], axis=1)[:, idx]
            var = self.var.reindex(var_slice)

        # obs slicing
        simple_flag = False
        if isinstance(obs_slice, slice):
            simple_flag = True
        else:  # iterable
            obs_slice = np.array(obs_slice)
            if obs_slice.ndim == 0:
                obs_slice = np.expand_dims(obs_slice, axis=0)
            if obs_slice.size == 0:
                obs_slice = obs_slice.astype(int)
                simple_flag = True
            if issubclass(obs_slice.dtype.type, np.bool_):
                obs_slice = np.where(obs_slice)[0]
            if issubclass(obs_slice.dtype.type, np.integer):
                simple_flag = True

        if simple_flag:
            exprs = exprs[obs_slice, :]
            obs = self.obs.iloc[obs_slice, :]
        else:  # Name slicing
            assert np.all(np.in1d(obs_slice, self.obs_names))
            extract_idx = np.vectorize(
                lambda x: np.where(self.obs_names == x)[0][0]
            )(obs_slice)
            exprs = exprs[extract_idx, :]
            obs = self.obs.reindex(obs_slice)

        return ExprDataSet(exprs=exprs, obs=obs, var=var, uns=self.uns)

    def clean_duplicate_vars(self) -> "ExprDataSet":
        r"""
        Clean up variables to preserve only the first occurrence of
        duplicated variables.

        Returns
        -------
        cleaned
            An ExprDataSet object with duplicated variables removed.
        """
        unique_vars, duplicate_mask = \
            set(), np.ones(self.var_names.size).astype(np.bool_)
        for idx, item in enumerate(self.var_names):
            if item in unique_vars:
                duplicate_mask[idx] = False
            else:
                unique_vars.add(item)
        return self[:, duplicate_mask]

    def find_variable_genes(
            self,
            x_low_cutoff: float = 0.1,
            x_high_cutoff: float = 8.0,
            y_low_cutoff: float = 1.0,
            y_high_cutoff: float = np.inf,
            num_bin: int = 20,
            binning_method: str = "equal_frequency",
            grouping: typing.Optional[str] = None,
            min_group_frac: float = 0.5,
    ) -> typing.Tuple[
        typing.List[str],
        typing.Union[matplotlib.axes.Axes, typing.Mapping[str, matplotlib.axes.Axes]]
    ]:
        r"""
        A reimplementation of the Seurat v2 "mean.var.plot" gene selection
        method in the "FindVariableGenes" function, with the extended ability
        of selecting variable genes within specified groups of cells and then
        combine results of individual groups. This is useful to minimize batch
        effect during feature selection.

        Parameters
        ----------
        x_low_cutoff
            Minimal log mean cutoff
        x_high_cutoff
            Maximal log mean cutoff
        y_low_cutoff
            Minimal log VMR cutoff
        y_high_cutoff
            Maximal log VMR cutoff
        num_bin
            Number of bins based on mean expression.
        binning_method
            How binning should be done based on mean expression.
            Available choices include {"equal_width", "equal_frequency"}.
        grouping
            Specify a column in the ``obs`` table that splits cells into
            several groups. Gene selection is performed in each group separately
            and results are combined afterwards.
        min_group_frac
            The minimal fraction of groups in which a gene must be selected
            for it to be kept in the final result.

        Returns
        -------
        list
            A list of selected variables
        ax
            VMR plot (a dict of plots if grouping is specified)
        """
        if grouping is not None:
            selected_dict, ax_dict = {}, {}
            groups = np.unique(self.obs[grouping])
            for group in groups:
                result = self[
                    self.obs[grouping] == group, :
                ].find_variable_genes(
                    x_low_cutoff=x_low_cutoff, x_high_cutoff=x_high_cutoff,
                    y_low_cutoff=y_low_cutoff, y_high_cutoff=y_high_cutoff,
                    num_bin=num_bin, binning_method=binning_method
                )
                selected_dict[group] = result[0]
                ax_dict[group] = result[1]
            selected = np.concatenate(list(selected_dict.values()))
            selected_unique, selected_count = np.unique(selected, return_counts=True)
            selected = selected_unique[selected_count >= min_group_frac * groups.size]
            return selected, ax_dict

        exprs = self.normalize().exprs
        mean = np.asarray(np.mean(exprs, axis=0)).ravel()
        var = np.asarray(np.mean(
            exprs.power(2) if scipy.sparse.issparse(exprs)
            else np.square(exprs),
            axis=0
        )).ravel() - np.square(mean)
        log_mean = np.log1p(mean)
        log_vmr = np.log(var / mean)
        log_vmr[np.isnan(log_vmr)] = 0
        if binning_method == "equal_width":
            log_mean_bin = pd.cut(log_mean, num_bin)
        elif binning_method == "equal_frequency":
            log_mean_bin = pd.cut(
                log_mean, [-1] + np.percentile(
                    log_mean[log_mean > 0], np.linspace(0, 100, num_bin)
                ).tolist()
            )
        else:
            raise ValueError("Invalid binning method!")
        summary_df = pd.DataFrame({
            "log_mean": log_mean,
            "log_vmr": log_vmr,
            "log_mean_bin": log_mean_bin
        }, index=self.var_names)
        summary_df["log_vmr_scaled"] = summary_df.loc[
            :, ["log_vmr", "log_mean_bin"]
        ].groupby("log_mean_bin").transform(lambda x: (x - x.mean()) / x.std())
        summary_df["log_vmr_scaled"].fillna(0, inplace=True)
        selected = summary_df.query(
            f"log_mean > {x_low_cutoff} & log_mean < {x_high_cutoff} & "
            f"log_vmr_scaled > {y_low_cutoff} & log_vmr_scaled < {y_high_cutoff}"
        )
        summary_df["selected"] = np.in1d(summary_df.index, selected.index)

        _, ax = plt.subplots(figsize=(7, 7))
        ax = sns.scatterplot(
            x="log_mean", y="log_vmr_scaled", hue="selected",
            data=summary_df, edgecolor=None, s=5, ax=ax
        )
        for _, row in selected.iterrows():
            ax.text(
                row["log_mean"], row["log_vmr_scaled"], row.name,
                size="x-small", ha="center", va="center"
            )
        ax.set_xlabel("Average expression")
        ax.set_ylabel("Dispersion")
        return selected.index.to_numpy().tolist(), ax

    def get_meta_or_var(
            self, names: typing.List[str], normalize_var: bool = False,
            log_var: bool = False
    ) -> pd.DataFrame:
        r"""
        Get either cell meta information (specified by column names in
        the ``obs`` table) or gene expression values in the expression matrix
        (specified by gene names).

        Parameters
        ----------
        names
            List of names that specifies meta information / genes to be fetched.
        normalize_var
            Whether to do cell-normalization before fetching variable values.
        log_var
            Whether to apply log transform for fetched variable values.

        Returns
        -------
        fetched
            Fetched result.
        """
        meta_names = np.intersect1d(names, self.obs.columns)
        remain_names = np.setdiff1d(names, self.obs.columns)
        var_names = np.intersect1d(remain_names, self.var_names)
        if var_names.size != remain_names.size:  # pragma: no cover
            raise ValueError("Unknown names encountered!")
        result = self.obs.loc[:, meta_names]
        if var_names.size:
            ds = self.normalize() if normalize_var else self
            exprs = utils.densify(ds[:, var_names].exprs)
            exprs = np.log1p(exprs) if log_var else exprs
            var = pd.DataFrame(exprs, index=self.obs_names, columns=var_names)
            result = pd.concat([result, var], axis=1)
        return result.loc[:, names]

    def copy(self, deep: bool = False) -> "ExprDataSet":
        r"""
        Produce a copy of the dataset.

        Parameters
        ----------
        deep
            Whether to perform deep copy.

        Returns
        -------
        copied
            Copy of the dataset.
        """
        if deep:
            return ExprDataSet(
                self.exprs.copy(), self.obs.copy(), self.var.copy(),
                copy.deepcopy(self.uns)
            )
        return ExprDataSet(self.exprs, self.obs, self.var, self.uns)

    def write_dataset(self, filename: str) -> None:
        r"""
        Write the dataset to a file.

        Parameters
        ----------
        filename
            File to be written (content in hdf5 format).
        """
        if not os.path.exists(os.path.dirname(filename)):
            os.makedirs(os.path.dirname(filename))
        h5opts = {**config.H5_COMPRESS_OPTS, **config.H5_TRACK_OPTS}
        with h5py.File(filename, "w") as f:
            if scipy.sparse.issparse(self.exprs):
                self.exprs.sort_indices()  # Compatibility with R
                g = f.create_group("exprs")
                g.create_dataset("data", data=self.exprs.data, **h5opts)
                g.create_dataset("indices", data=self.exprs.indices, **h5opts)
                g.create_dataset("indptr", data=self.exprs.indptr, **h5opts)
                g.create_dataset("shape", data=self.exprs.shape, **h5opts)
            else:
                f.create_dataset("exprs", data=self.exprs, **h5opts)
            f.create_dataset("obs_names", data=utils.encode(self.obs_names.values), **h5opts)
            f.create_dataset("var_names", data=utils.encode(self.var_names.values), **h5opts)
            dict_to_group(df_to_dict(self.obs), f.create_group("obs"))
            dict_to_group(df_to_dict(self.var), f.create_group("var"))
            dict_to_group(self.uns, f.create_group("uns"))

    @classmethod
    def read_dataset(
            cls, filename: str, sparsify: bool = False,
            skip_exprs: bool = False
    ) -> "ExprDataSet":
        r"""
        Read dataset from file (saved by ``write_dataset``).

        Parameters
        ----------
        filename
            File to read from (content in hdf5 format).
        sparsify
            Whether to coerce the expression matrix into sparse format.
        skip_exprs
            Whether to skip reading the expression matrix (fill with all
            zeros instead). This option is for accelerating data reading
            in case if only the meta information are needed.

        Returns
        -------
        loaded_dataset
            An ExprDataSet object loaded from the file.
        """
        with h5py.File(filename, "r") as f:
            obs = pd.DataFrame(
                dict_from_group(f["obs"]),
                index=utils.decode(f["obs_names"][...])
            )
            var = pd.DataFrame(
                dict_from_group(f["var"]),
                index=utils.decode(f["var_names"][...])
            )
            uns = dict_from_group(f["uns"])

            if not skip_exprs:
                exprs_handle = f["exprs"]
                if isinstance(exprs_handle, h5py.Group):  # Sparse matrix
                    mat = scipy.sparse.csr_matrix((
                        exprs_handle['data'][...],
                        exprs_handle['indices'][...],
                        exprs_handle['indptr'][...]
                    ), shape=exprs_handle['shape'][...])
                else:  # Dense matrix
                    mat = exprs_handle[...].astype(np.float32)
                    if sparsify:
                        mat = scipy.sparse.csr_matrix(mat)
            else:
                mat = scipy.sparse.csr_matrix((obs.shape[0], var.shape[0]))
        return cls(exprs=mat, obs=obs, var=var, uns=uns)

    def map_vars(
            self, mapping: pd.DataFrame,
            map_uns_slots: typing.Optional[typing.List[str]] = None
    ) -> "ExprDataSet":
        r"""
        Map variables of the dataset to some other terms,
        e.g. gene ortholog groups, or orthologous genes in another species.

        Parameters
        ----------
        mapping
            A 2-column data frame defining variable name mapping. First column
            is source variable name and second column is target variable name.
        map_uns_slots
            Assuming variable subsets, e.g. most informative genes,
            are stored in the ``uns`` slot, this parameter specifies which slots
            in ``uns`` should also be mapped.
            Note that ``uns`` slots not specified here will be left as is.

        Returns
        -------
        mapped
            Mapped dataset.
        """
        # Convert to mapping matrix
        source = self.var_names
        mapping = mapping.loc[np.in1d(mapping.iloc[:, 0], source), :]
        target = np.unique(mapping.iloc[:, 1])

        source_idx_map = {val: i for i, val in enumerate(source)}
        target_idx_map = {val: i for i, val in enumerate(target)}
        source_idx = [source_idx_map[val] for val in mapping.iloc[:, 0]]
        target_idx = [target_idx_map[val] for val in mapping.iloc[:, 1]]
        mapping = scipy.sparse.csc_matrix(
            (np.repeat(1, mapping.shape[0]), (source_idx, target_idx)),
            shape=(source.size, target.size)
        )

        # Sanity check
        amb_src_mask = np.asarray(mapping.sum(axis=1)).squeeze() > 1
        amb_tgt_mask = np.asarray(mapping.sum(axis=0)).squeeze() > 1
        if amb_src_mask.sum() > 0:  # pragma: no cover
            utils.logger.warning("%d ambiguous source items found!",
                                 amb_src_mask.sum())
            utils.logger.info(str(source[amb_src_mask].tolist()))
        if amb_tgt_mask.sum() > 0:  # pragma: no cover
            utils.logger.warning("%d ambiguous target items found!",
                                 amb_tgt_mask.sum())
            utils.logger.info(str(target[amb_tgt_mask].tolist()))

        # Compute new expression matrix
        new_exprs = self.exprs @ mapping
        if scipy.sparse.issparse(self.exprs):
            new_exprs = scipy.sparse.csr_matrix(new_exprs)

        # Update var accordingly
        new_var = pd.DataFrame(index=target)

        # Update uns accordingly
        if map_uns_slots is None:
            map_uns_slots = []
        new_uns = {}
        for slot in self.uns:
            if slot in map_uns_slots:
                idx = [source_idx_map[val] for val in self.uns[slot]]
                idx = np.where(np.asarray(
                    mapping[idx, :].sum(axis=0)
                ).squeeze() > 0)[0]
                new_uns[slot] = target[idx]
            else:
                new_uns[slot] = self.uns[slot]
        return ExprDataSet(new_exprs, self.obs, new_var, new_uns)

    @classmethod
    def merge_datasets(
            cls, dataset_dict: typing.Mapping[str, "ExprDataSet"],
            meta_col: typing.Optional[str] = None,
            merge_uns_slots: typing.Optional[typing.List[str]] = None
    ) -> "ExprDataSet":
        r"""
        Merge multiple dataset objects into a single "meta-dataset".

        Parameters
        ----------
        dataset_dict
            A dict of ExprDataSet objects. Dict keys will be used as values in
            ``meta_col`` (see ``meta_col``).
        meta_col
            Name of a new column to be added to ``obs`` table of the merged
            ExprDataSet object, which can be used for distinguishing cells from
            each dataset. If not specified, no such column will be added.
        merge_uns_slots
            Assuming variable subsets, e.g. most informative genes,
            are stored in the ``uns`` slot, this parameter specifies the
            variable subsets to be merged.
            Note that ``uns`` slots not specified here will be discarded.

        Returns
        -------
        merged_dataset
            Merged dataset.
        """
        dataset_dict = collections.OrderedDict(dataset_dict)

        var_name_list = [dataset.var_names for dataset in dataset_dict.values()]
        var_union = functools.reduce(np.union1d, var_name_list)
        var_intersect = functools.reduce(np.intersect1d, var_name_list)

        for item in dataset_dict:
            dataset_dict[item] = dataset_dict[item].copy(deep=True)[
                :, var_union
            ]  # Avoid contaminating original datasets

        utils.logger.info("Merging uns slots...")
        if merge_uns_slots is None:
            merge_uns_slots = []
        merged_slot = {}
        for slot in merge_uns_slots:
            merged_slot[slot] = []
            for dataset in dataset_dict.values():
                merged_slot[slot].append(dataset.uns[slot])
            merged_slot[slot] = np.intersect1d(
                functools.reduce(np.union1d, merged_slot[slot]), var_intersect)

        utils.logger.info("Merging var data frame...")
        merged_var = []
        for item in dataset_dict:
            var = dataset_dict[item].var.reindex(var_union)
            var.columns = ["_".join([c, item]) for c in var.columns]
            merged_var.append(var)
        merged_var = pd.concat(merged_var, axis=1)

        utils.logger.info("Merging obs data frame...")
        merged_obs = []
        for key in dataset_dict.keys():
            if meta_col:
                dataset_dict[key].obs[meta_col] = key
            merged_obs.append(dataset_dict[key].obs)
        merged_obs = pd.concat(merged_obs, sort=True)

        utils.logger.info("Merging expression matrix...")
        if np.any([
                scipy.sparse.issparse(dataset.exprs)
                for dataset in dataset_dict.values()
        ]):
            merged_exprs = scipy.sparse.vstack([
                scipy.sparse.csr_matrix(dataset.exprs)
                for dataset in dataset_dict.values()
            ])
        else:
            merged_exprs = np.concatenate([
                dataset.exprs for dataset in dataset_dict.values()
            ], axis=0)

        return cls(
            exprs=merged_exprs, obs=merged_obs, var=merged_var,
            uns=merged_slot
        )

    def _prepare_latent_visualization(
            self, method: str, random_seed: int = config._USE_GLOBAL,
            reuse: bool = True, **kwargs
    ) -> None:
        import sklearn.manifold
        import umap

        random_seed = config.RANDOM_SEED \
            if random_seed == config._USE_GLOBAL else random_seed

        # method in {"tSNE", "UMAP"}
        if not reuse:
            for col in self.obs.columns:
                if col.startswith(method):
                    del self.obs[col]
        mask = np.vectorize(lambda x: x.startswith(method))(self.obs.columns)
        if not np.any(mask):
            if method == "tSNE":
                mapper = sklearn.manifold.TSNE(random_state=random_seed, **kwargs)
            elif method == "UMAP":
                mapper = umap.UMAP(random_state=random_seed, **kwargs)
            else:
                raise ValueError("Unknown method!")
            utils.logger.info("Computing %s...", method)
            coord = mapper.fit_transform(self.latent)
            columns = np.vectorize(
                lambda x, method=method: f"{method}{x}"
            )(np.arange(coord.shape[1]) + 1)
            coord_df = pd.DataFrame(coord, index=self.obs_names, columns=columns)
            self.obs = pd.concat([self.obs, coord_df], axis=1)
        else:
            utils.logger.info("Using cached %s...", method)

    def visualize_latent(
            self, hue: typing.Optional[str] = None,
            style: typing.Optional[str] = None,
            method: str = "tSNE", reuse: bool = True, shuffle: bool = True,
            sort: bool = False, ascending: bool = True, size: float = 3.0,
            width: float = 7.0, height: float = 7.0,
            random_seed: int = config._USE_GLOBAL,
            ax: typing.Optional[matplotlib.axes.Axes] = None,
            dr_kws: typing.Optional[typing.Mapping] = None,
            scatter_kws: typing.Optional[typing.Mapping] = None
    ) -> matplotlib.axes.Axes:
        r"""
        Visualize latent space

        Parameters
        ----------
        hue
            Specifies a column in the ``obs`` table or a gene name to use as
            hue of the data points.
        style
            Specifies a column in the ``obs`` table to use as style of data
            points.
        method
            Should be among {"tSNE", "UMAP", None}.
            Specifies the dimension reduction algorithm for visualization.
            If ``None`` is specified, the first two latent
            dimensions will be used for visualization.
        reuse
            Whether to reuse existing visualization coordinates.
        shuffle
            Whether to shuffle data points before plotting.
        sort
            Whether to sort points according to ``hue`` before plotting.
            If set to true, ``shuffle`` takes no effect.
        ascending
            Whether sorting is in the ascending order.
            Only effective when ``sort`` is set to true.
        size
            Point size.
        width
            Figure width.
        height
            Figure height.
        random_seed
            Random seed used in dimension reduction algorithm. If not specified,
            :data:`config.RANDOM_SEED` will be used, which defaults
            to None.
        ax
            Specifies an existing axes to plot onto. If specified,
            ``width`` and ``height`` take no effect.
        dr_kws
            Keyword arguments passed to the dimension reduction algorithm,
            according to ``method``.
            If ``method`` is "tSNE", will be passed to :class:`sklearn.manifold.TSNE`.
            If ``method`` is "UMAP", will be passed to :class:`umap.UMAP`.
        scatter_kws
            Keyword arguments to be passed to :func:`sns.scatterplot`.

        Returns
        -------
        ax
            Visualization plot.
        """
        if dr_kws is None:
            dr_kws = dict()
        if scatter_kws is None:
            scatter_kws = dict()

        random_seed = config.RANDOM_SEED \
            if random_seed == config._USE_GLOBAL else random_seed

        if method is not None:
            self._prepare_latent_visualization(
                method, random_seed=random_seed, reuse=reuse, **dr_kws)
        else:
            method = "latent_"
        if ax is None:
            _, ax = plt.subplots(figsize=(width, height))
        fetch = [f"{method}1", f"{method}2"]
        if hue is not None:
            fetch.append(hue)
        if style is not None:
            fetch.append(style)
        df = self.get_meta_or_var(fetch, normalize_var=True, log_var=True)
        if shuffle:
            df = df.sample(frac=1, random_state=random_seed)
        if hue is not None and sort:
            df = df.sort_values(hue, ascending=ascending)
        ax = sns.scatterplot(
            x=f"{method}1", y=f"{method}2",
            hue=hue, style=style, s=size, data=df, edgecolor=None, ax=ax,
            **scatter_kws
        )
        ax.spines["right"].set_visible(False)
        ax.spines["top"].set_visible(False)
        ax.yaxis.set_ticks_position("left")
        ax.xaxis.set_ticks_position("bottom")
        if hue is not None:
            _ = ax.legend(
                bbox_to_anchor=(1.05, 0.5), loc="center left",
                borderaxespad=0.0, frameon=False
            )
        return ax

    def obs_correlation_heatmap(
            self, group: typing.Optional[str] = None,
            used_vars: typing.Optional[typing.List[str]] = None,
            cluster_method: str = "complete",
            width: float = 10.0, height: float = 10.0, **kwargs
    ) -> sns.matrix.ClusterGrid:
        r"""
        Correlation heatmap of each observation.

        Parameters
        ----------
        group
            Specifies a column in the ``obs`` table which will be used to label
            rows and columns.
        used_vars
            Specifies variables used to compute correlation. If not specified,
            meaning all variables will be used.
        cluster_method
            Clustering method. See :func:``scipy.cluster.hierarchy.linkage``
            for available options.
        width
            Figure width.
        height
            Figure height.
        kwargs
            Additional keyword arguments will be passed to
            :func:`sns.clustermap`.

        Returns
        -------
        grid
            Visualization plot.
        """
        import matplotlib.patches as mpatches
        import scipy.cluster

        dataset = self if used_vars is None else self[:, used_vars]
        exprs = utils.densify(dataset.exprs)
        exprs = np.log1p(exprs)
        corr = pd.DataFrame(
            np.corrcoef(exprs),
            index=dataset.obs_names, columns=dataset.obs_names
        )
        linkage = scipy.cluster.hierarchy.linkage(
            exprs, method=cluster_method, metric="correlation")
        if group is not None:
            label = self.obs[group]
            label_uniq = np.unique(label.values)
            lut = collections.OrderedDict(zip(label_uniq, sns.hls_palette(
                label_uniq.size, l=0.5, s=0.7)))
            legend_patch = [mpatches.Patch(color=lut[k], label=k) for k in lut]
            label = label.map(lut)
        else:
            legend_patch = None
            label = None
        grid = sns.clustermap(
            corr, row_linkage=linkage, col_linkage=linkage,
            row_colors=label, col_colors=label, figsize=(width, height),
            xticklabels=False, yticklabels=False, **kwargs
        )
        if legend_patch is not None:
            _ = grid.ax_heatmap.legend(
                loc="center left", bbox_to_anchor=(1.05, 0.5),
                handles=legend_patch, frameon=False
            )
        return grid

    def violin(
            self, group: str, var: str, normalize_var: bool = True,
            width: float = 7, height: float = 7,
            ax: typing.Optional[matplotlib.axes.Axes] = None,
            strip_kws: typing.Optional[typing.Mapping] = None,
            violin_kws: typing.Optional[typing.Mapping] = None
    ) -> matplotlib.axes.Axes:
        r"""
        Violin plot across obs groups.

        Parameters
        ----------
        group
            Specifies a column in the ``obs`` table used for cell grouping.
        var
            Variable name.
        normalize_var
            Whether to perform cell normalization.
        width
            Figure width.
        height
            Figure height.
        ax
            Specifies an existing axes to plot onto. If specified,
            ``width`` and ``height`` take no effect.
        strip_kws
            Additional keyword arguments will be passed to :func:`sns.stripplot`.
        violin_kws
            Additional keyword arguments will be passed to :func:`sns.violinplot`.

        Returns
        -------
        ax
            Visualization figure.
        """
        strip_kws = strip_kws or {}
        violin_kws = violin_kws or {}

        df = self.get_meta_or_var(
            [group, var],
            normalize_var=normalize_var, log_var=True
        )
        if ax is None:
            _, ax = plt.subplots(figsize=(width, height))
        ax = sns.stripplot(
            x=group, y=var, data=df,
            color=".3", edgecolor=None, size=3, ax=ax, **strip_kws
        )
        ax = sns.violinplot(
            x=group, y=var, data=df,
            scale="width", ax=ax, inner=None, **violin_kws
        )
        ax.spines["right"].set_visible(False)
        ax.spines["top"].set_visible(False)
        return ax

    def annotation_confidence(
            self, annotation: typing.Union[str, typing.List[str]],
            used_vars: typing.Optional[typing.Union[str, typing.List[str]]] = None,
            metric: str = "cosine", return_group_percentile: bool = True
    ) -> typing.Tuple[np.ndarray, np.ndarray]:
        r"""
        Compute annotation confidence of each obs (cell) based on
        sample silhouette score.

        Parameters
        ----------
        annotation
            Specifies annotation for which confidence will be computed.
            If passed an array-like, it should be 1 dimensional with length
            equal to obs number, and will be used directly as annotation.
            If passed a string, it should be a column name in ``obs``.
        used_vars
            Specifies the variable set used to evaluate ``metric``,
            If not specified, all variables are used. If given a string,
            it should be a slot in `uns`. If given a 1-d array, it should
            contain variable names to be used.
        metric
            Specifies distance metric used to compute sample silhouette scores.
            See :func:`sklearn.metrics.silhouette_samples` for available
            options.
        return_group_percentile
            Whether to return within group confidence percentile, instead of
            raw sample silhouette score.

        Returns
        -------
        confidence
            1 dimensional numpy array containing annotation confidence for
            each obs.
        group_percentile
            1 dimensional numpy array containing within-group percentile
            for each obs.
        """
        if isinstance(annotation, str):
            annotation = self.obs[annotation].values
        annotation = utils.encode_integer(annotation)[0]
        if used_vars is None:
            used_vars = self.var_names
        elif isinstance(used_vars, str):
            used_vars = self.uns[used_vars]
        exprs = self[:, used_vars].exprs
        exprs = exprs.log1p() if scipy.sparse.issparse(exprs) \
            else np.log1p(exprs)
        confidence = sklearn.metrics.silhouette_samples(
            exprs, annotation, metric=metric)
        if return_group_percentile:
            normalized_confidence = np.zeros_like(confidence)
            for l in np.unique(annotation):
                mask = annotation == l
                normalized_confidence[mask] = (
                    scipy.stats.rankdata(confidence[mask]) - 1
                ) / (mask.sum() - 1)
            return confidence, normalized_confidence
        return confidence

    def fast_markers(
            self, group: str,
            used_genes: typing.Optional[typing.List[str]] = None,
            alternative: str = "two-sided", multitest: str = "bonferroni",
            min_pct: float = 0.1, min_pct_diff: float = -np.inf,
            logfc_threshold: float = 0.25, pseudocount: float = 1.0,
            n_jobs: int = 1
    ) -> typing.Mapping[str, pd.DataFrame]:
        r"""
        Find markers for each group by one-vs-rest Wilcoxon rank sum test.
        This is a fast implementation of the ``FindAllMarkers`` function
        in Seurat v2.

        Parameters
        ----------
        group
            Specifies a column in ``obs`` that determines cell grouping.
        used_genes
            A sequence of genes in which to search for markers.
        alternative
            Alternative hypothesis, should be among
            {"two-sided", "greater", "less"}.
        multitest
            Method of multiple test p-value correction. Check
            :func:`statsmodels.stats.multitest.multipletests` for available
            options.
        min_pct
            Minimal percent of cell expressing gene of interest, either in
            group or rest, for it to be considered in statistical test.
        min_pct_diff
            Minimal percent difference of cell expressing gene of interest
            in group and rest, for it to be considered in statistical test.
        logfc_threshold
            Minimal log fold change in average expression level of gene
            of interest, between group and rest, for it to be considered in
            statistical test.
        pseudocount
            Pseudocount to be added when computing log fold change.
        n_jobs
            Number of parallel running threads to use.

        Returns
        -------
        summary
            Each element, named by cell group, is a table containing
            differential expression results.
            Columns of each table are:
            "pct_1": percent of cells expressing the gene in group of interest
            "pct_2": percent of cells expressing the gene in rest
            "logfc": log fold change of mean expression between group and rest
            "stat": statistic of Wilcoxon rank sum test
            "z": normal approximation statistic of Wilcoxon rank sum test
            "pval": p-value of Wilcoxon rank sum test
            "padj": p-value adjusted for multiple test
        """
        import statsmodels.stats.multitest

        if used_genes is None:
            used_genes = self.var_names
        mat = self[:, used_genes].exprs.toarray()
        group_indices, labels = utils.encode_integer(self.obs[group].values)
        group_onehot = utils.encode_onehot(group_indices).astype(bool).toarray()
        n_x = np.sum(group_onehot, axis=0).astype(np.float64)
        n_y = group_onehot.shape[0] - n_x
        n_xy_prod = n_x * n_y
        n_xy_plus = n_x + n_y

        def ranksum_thread(vec):
            r"""
            Wilcoxon rank sum test for one feature
            Adapted from R functions:
                ``Seurat::FindMarkers`` and ``stats::wilcox.test``
            """

            # Preparation
            vec = vec.toarray().ravel() if scipy.sparse.issparse(vec) else vec
            pct = np.empty((group_onehot.shape[1], 2))
            logfc = np.empty((group_onehot.shape[1]))

            for i in range(group_onehot.shape[1]):
                mask = group_onehot[:, i].ravel()
                pct[i, 0] = round(np.sum(vec[mask] > 0) / n_x[i], 3)
                pct[i, 1] = round(np.sum(vec[~mask] > 0) / n_y[i], 3)
                logfc[i] = np.log(vec[mask].mean() + pseudocount) - \
                    np.log(vec[~mask].mean() + pseudocount)

            # Percent expressed filtering
            pct_max = pct.max(axis=1)
            pct_min = pct.min(axis=1)
            pct_diff = pct_max - pct_min
            pct_mask = (pct_max > min_pct) & (pct_diff > min_pct_diff)

            # Fold change filtering
            if alternative == "greater":
                logfc_mask = logfc > logfc_threshold
            elif alternative == "less":
                logfc_mask = logfc < -logfc_threshold
            elif alternative == "two-sided":
                logfc_mask = abs(logfc) > logfc_threshold

            total_mask = pct_mask & logfc_mask
            if not np.any(total_mask):
                nan_placeholder = np.empty(group_onehot.shape[1])
                nan_placeholder[:] = np.nan
                return pct[:, 0].ravel(), pct[:, 1].ravel(), logfc, \
                    nan_placeholder, nan_placeholder, nan_placeholder

            # Rank sum test
            rank = scipy.stats.rankdata(vec)
            n_ties = np.unique(rank, return_counts=True)[1]

            stat = np.empty(group_onehot.shape[1])
            for i in range(group_onehot.shape[1]):
                mask = group_onehot[:, i].ravel()
                if total_mask[i]:
                    stat[i] = rank[mask].sum() - n_x[i] * (n_x[i] + 1) / 2
                else:
                    stat[i] = np.nan
            z = stat - n_x * n_y / 2
            sigma = np.sqrt((n_xy_prod / 12) * (
                (n_xy_plus + 1) -
                (n_ties ** 3 - n_ties).sum() / (n_xy_plus * (n_xy_plus - 1))
            ))
            if alternative == "two-sided":
                correction = np.sign(z) * 0.5
            elif alternative == "greater":
                correction = 0.5
            elif alternative == "less":
                correction = -0.5
            z = (z - correction) / sigma
            if alternative == "two-sided":
                pval = 2 * np.stack([
                    scipy.stats.norm.sf(z),
                    scipy.stats.norm.cdf(z)
                ], axis=0).min(axis=0)
            elif alternative == "greater":
                pval = scipy.stats.norm.sf(z)
            elif alternative == "less":
                pval = scipy.stats.norm.cdf(z)
            return pct[:, 0].ravel(), pct[:, 1].ravel(), logfc, stat, z, pval

        with concurrent.futures.ThreadPoolExecutor(max_workers=n_jobs) as executor:
            result = executor.map(ranksum_thread, mat.T)

        pct_1_list, pct_2_list, logfc_list, stat_list, z_list, pval_list = \
            [], [], [], [], [], []
        for pct_1, pct_2, logfc, stat, z, pval in result:
            pct_1_list.append(pct_1)
            pct_2_list.append(pct_2)
            logfc_list.append(logfc)
            stat_list.append(stat)
            z_list.append(z)
            pval_list.append(pval)
        pct_1 = np.stack(pct_1_list, axis=0).T
        pct_2 = np.stack(pct_2_list, axis=0).T
        logfc = np.stack(logfc_list, axis=0).T
        stat = np.stack(stat_list, axis=0).T
        z = np.stack(z_list, axis=0).T
        pval = np.stack(pval_list, axis=0).T

        def my_multitest(pval):
            pval = pval.copy()
            mask = ~np.isnan(pval)
            if np.any(mask):
                pval[mask] = statsmodels.stats.multitest.multipletests(
                    pval[mask], method=multitest
                )[1]
            return pval
        padj = np.apply_along_axis(my_multitest, 1, pval)

        summary = {}
        for i in range(group_onehot.shape[1]):
            summary[labels[i]] = pd.DataFrame({
                "pct_1": pct_1[i],
                "pct_2": pct_2[i],
                "logfc": logfc[i],
                "stat": stat[i],
                "z": z[i],
                "pval": pval[i],
                "padj": padj[i]
            }, index=used_genes).sort_values(by=["z"], ascending=False)
        return summary

    def to_anndata(self) -> anndata.AnnData:
        r"""
        Convert a :class:`ExprDataSet` object to an
        :class:`anndata.AnnData` object.

        Returns
        -------
        dataset
            Resulting :class:`anndata.AnnData` object.
        """
        return anndata.AnnData(
            X=self.exprs, obs=self.obs, var=self.var, uns=dict(self.uns))

    @classmethod
    def from_anndata(cls, ad: anndata.AnnData) -> "ExprDataSet":
        r"""
        Create a :class:`ExprDataSet` object from an existing
        :class:`anndata.AnnData` object.

        Parameters
        ----------
        ad
            An existing :class:`anndata.AnnData` object.

        Returns
        -------
        dataset
            Resulting :class:`ExprDataSet` object.
        """
        return cls(ad.X, ad.obs, ad.var, ad.uns)

    def to_loom(self, file: str) -> loompy.loompy.LoomConnection:
        r"""
        Convert a :class:`ExprDataSet` object to a
        :class:`loompy.loompy.LoomConnection` object. Note that data will be
        written to a loom file specified by ``file`` in this process.

        Parameters
        ----------
        file
            Specifies the loom file to be written

        Returns
        -------
        lm
            Resulting connection to the loom file.
        """
        assert "var_name" not in self.var.columns and \
            "obs_name" not in self.obs.columns
        loompy.create(file, self.exprs.T, {**{
            key: np.array(val)
            for key, val in self.var.to_dict("list").items()
        }, "var_name": np.array(self.var.index)}, {**{
            key: np.array(val)
            for key, val in self.obs.to_dict("list").items()
        }, "obs_name": np.array(self.obs.index)})
        return loompy.connect(file)

    @classmethod
    def from_loom(cls, lm: loompy.loompy.LoomConnection) -> "ExprDataSet":
        r"""
        Create a :class:`ExprDataSet` object from an existing
        :class:`loompy.loompy.LoomConnection` object.

        Parameters
        ----------
        lm
            An existing :class:`loompy.loompy.LoomConnection` object.

        Returns
        -------
        dataset
            Resulting :class:`ExprDataSet` object.
        """
        return cls(
            lm[:, :].T,
            pd.DataFrame({
                key: val for key, val in lm.ca.items()
                if key != "obs_name"
            }, index=lm.ca["obs_name"] if "obs_name" in lm.ca.keys() else None),
            pd.DataFrame({
                key: val for key, val in lm.ra.items()
                if key != "var_name"
            }, index=lm.ra["var_name"] if "var_name" in lm.ra.keys() else None),
            {}
        )

    def write_table(self, filename: str, orientation: str = "cg", **kwargs) -> None:
        r"""
        Write the expression matrix to a plain-text file.
        Note that ``obs`` (cell) meta table, ``var`` (gene) meta table and data
        in the ``uns`` slot are discarded, only the expression matrix is written
        to the file.

        Parameters
        ----------
        filename
            Name of the file to be written.
        orientation
            Specifies whether to write in :math:`obs \times var` or
            :math:`obs \times var` orientation, should be among {"cg", "gc"}.
        kwargs
            Additional keyword arguments will be passed to
            :meth:`pandas.DataFrame.to_csv`.
        """
        if not os.path.exists(os.path.dirname(filename)):
            os.makedirs(os.path.dirname(filename))
        if orientation == "cg":
            df = pd.DataFrame(
                utils.densify(self.exprs),
                index=self.obs_names,
                columns=self.var_names
            )
        elif orientation == "gc":
            df = pd.DataFrame(
                utils.densify(self.exprs.T),
                index=self.var_names,
                columns=self.obs_names
            )
        else:  # pragma: no cover
            raise ValueError("Invalid orientation!")
        df.to_csv(filename, **kwargs)

    @classmethod
    def read_table(
            cls, filename: str, orientation: str = "cg",
            sparsify: bool = False, **kwargs
    ) -> "ExprDataSet":
        r"""
        Read expression matrix from a plain-text file

        Parameters
        ----------
        filename
            Name of the file to read from.
        orientation
            Specifies whether matrix in the file is in
            :math:`cell \times gene` or :math:`gene \times cell` orientation.
        sparsify
            Whether to convert the expression matrix into sparse format.
        kwargs
            Additional keyword arguments will be passed to :func:`pandas.read_csv`.

        Returns
        -------
        loaded_dataset
            An :class:`ExprDataSet` object loaded from the file.
        """
        df = pd.read_csv(filename, **kwargs)
        if orientation == "gc":
            df = df.T
        return cls(
            scipy.sparse.csr_matrix(df.values) if sparsify else df.values,
            pd.DataFrame(index=df.index),
            pd.DataFrame(index=df.columns),
            {}
        )


def read_clean(data):
    assert isinstance(data, np.ndarray)
    if data.dtype.type is np.bytes_:
        data = utils.decode(data)
        mask = data == config._NAN_REPLACEMENT
        if np.any(mask):
            data = data.astype(object)
            data[mask] = np.nan
    if data.size == 1:
        data = data.flat[0]
    return data


def write_clean(data):
    if not isinstance(data, np.ndarray):
        data = np.array(data)
    if data.dtype.type is np.object_:
        data[utils.isnan(data)] = config._NAN_REPLACEMENT
    if data.dtype.type in (np.str_, np.object_):
        data = utils.encode(data)
    return data


def dict_from_group(group):
    assert isinstance(group, h5py.Group)
    # d = utils.dotdict()
    d = {}
    for key in group:
        if isinstance(group[key], h5py.Group):
            value = dict_from_group(group[key])
        else:
            value = read_clean(group[key][...])
        d[key] = value
    return d


def dict_to_group(d, group):
    for key in d:
        if isinstance(d[key], dict):
            dict_to_group(d[key], group.create_group(key))
        else:
            try:
                value = write_clean(d[key])
                if value.size == 1:
                    h5opts = config.H5_TRACK_OPTS
                else:
                    h5opts = {**config.H5_COMPRESS_OPTS, **config.H5_TRACK_OPTS}
                group.create_dataset(key, data=value, **h5opts)
            except Exception:  # pylint: disable=broad-except
                utils.logger.warning("Slot %s failed to save!", key)


def df_to_dict(df):
    d = collections.OrderedDict()
    for column in df.columns:
        d[column] = df[column].values
    return d


def check_hybrid_path(hybrid_path):
    file_name, h5_path = hybrid_path.split("//")
    if not os.path.exists(file_name):
        return False
    with h5py.File(file_name, "r") as f:
        return h5_path in f


def read_hybrid_path(hybrid_path):
    file_name, h5_path = hybrid_path.split("//")
    with h5py.File(file_name, "r") as f:
        if isinstance(f[h5_path], h5py.Group):
            fetch = dict_from_group(f[h5_path])
        else:
            fetch = read_clean(f[h5_path][...])
    return fetch


def write_hybrid_path(obj, hybrid_path):
    file_name, h5_path = hybrid_path.split("//")
    if not os.path.exists(os.path.dirname(file_name)):
        os.makedirs(os.path.dirname(file_name))
    with h5py.File(file_name, "a") as f:
        if h5_path in f:
            del f[h5_path]
        if isinstance(obj, dict):
            dict_to_group(obj, f.create_group(h5_path))
        else:
            obj = write_clean(obj)
            if obj.size == 1:
                h5opts = config.H5_TRACK_OPTS
            else:
                h5opts = {**config.H5_COMPRESS_OPTS, **config.H5_TRACK_OPTS}
            f.create_dataset(h5_path, data=obj, **h5opts)



import sys
import os
import shutil
import unittest
import numpy as np
import matplotlib
matplotlib.use("agg")

if os.environ.get("TEST_MODE", "INSTALL") == "DEV":
    sys.path.insert(0, "..")
import Cell_BLAST as cb
cb.config.RANDOM_SEED = 0
cb.config.N_JOBS = 2

os.environ["TF_CPP_MIN_LOG_LEVEL"] = "2"


class BLASTTest(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.data = cb.data.ExprDataSet.read_dataset(
            "pollen.h5"
        ).normalize()
        cls.models = []
        for i in range(2):
            cls.models.append(cb.directi.fit_DIRECTi(
                cls.data, cls.data.uns["scmap_genes"], supervision="cell_type1",
                latent_dim=10, epoch=3, random_seed=i, path="test_directi"
            ))

    @classmethod
    def tearDownClass(cls):
        if os.path.exists("./test_directi"):
            shutil.rmtree("./test_directi")

    def tearDown(self):
        if os.path.exists("./test_blast"):
            shutil.rmtree("./test_blast")
        if os.path.exists("./tmp"):
            shutil.rmtree("./tmp")

    def test_npdv1_dist_blast(self):
        blast = cb.blast.BLAST(self.models, self.data, distance_metric="npd_v1", n_posterior=20, force_components=False)
        hits = blast[:].query(self.data)
        self.assertEqual(len(hits), self.data.shape[0])
        prediction = hits.reconcile_models(dist_method="min").filter(by="dist").annotate("cell_type1", return_evidence=True).loc[:, ["cell_type1", "n_hits"]]
        _ = hits.reconcile_models().to_data_frames()
        blast.save("./test_blast")
        blast2 = cb.blast.BLAST.load("./test_blast")
        hits2 = blast2.query(self.data)
        prediction2 = hits2.reconcile_models(dist_method="min").filter(by="dist").annotate("cell_type1", return_evidence=True).loc[:, ["cell_type1", "n_hits"]]
        self.assertTrue(np.all(prediction.values == prediction2.values))

    def test_euclidean_dist_blast(self):
        blast = cb.blast.BLAST(self.models, self.data, distance_metric="ed", cluster_empirical=True)
        hits = blast.query(self.data)
        prediction = hits.reconcile_models(dist_method="max").filter(by="dist").annotate("cell_type1")
        blast.save("./test_blast")
        blast2 = cb.blast.BLAST.load("./test_blast")
        hits2 = blast2.query(self.data)
        prediction2 = hits2.reconcile_models(dist_method="max").filter(by="dist").annotate("cell_type1")
        self.assertTrue(np.all(prediction.values == prediction2.values))
        _ = hits.filter(by="dist")[0]
        _ = hits.reconcile_models().filter(by="dist")[0:3]

    def test_amd_pval_blast(self):
        blast = cb.blast.BLAST(self.models, self.data, distance_metric="amd", n_posterior=20, cluster_empirical=True, force_components=False)
        hits = blast.query(self.data, store_dataset=True)
        prediction = hits.filter(by="pval", cutoff=0.05).annotate("cell_type1")
        blast.save("./test_blast")
        blast2 = cb.blast.BLAST.load("./test_blast")
        hits2 = blast2.query(self.data)
        prediction2 = hits2.filter(by="pval", cutoff=0.05).annotate("cell_type1")
        self.assertTrue(np.all(prediction.values == prediction2.values))
        _ = hits.filter(by="pval", cutoff=0.05)[0]
        _ = hits.reconcile_models().filter(by="pval", cutoff=0.05)[0:3]

    def test_npdv2_pval_blast(self):
        blast = cb.blast.BLAST(self.models, self.data, distance_metric="npd_v2", n_posterior=20)
        hits = blast.query(self.data)
        for _ in hits:
            pass
        prediction = hits.reconcile_models().filter(by="pval").annotate("ABRACL").fillna(0)
        blast.save("./test_blast")
        blast2 = cb.blast.BLAST.load("./test_blast")
        hits2 = blast2.query(self.data, store_dataset=True)
        prediction2 = hits2.reconcile_models().filter(by="pval").annotate("ABRACL").fillna(0)
        self.assertTrue(np.all(prediction.values == prediction2.values))

    def test_align_blast(self):
        blast = cb.blast.BLAST(self.models, self.data, distance_metric="md", n_posterior=20, force_components=False)
        blast = blast.align(dict(query=self.data), path="tmp", epoch=3, patience=3)
        hits = blast.query(self.data)
        prediction = hits.reconcile_models().filter(by="pval").annotate("cell_type1")
        blast.save("./test_blast")
        blast2 = cb.blast.BLAST.load("./test_blast", mode=cb.blast.MINIMAL)
        hits2 = blast2.query(self.data, store_dataset=True)
        prediction2 = hits2.reconcile_models().filter(by="pval").annotate("cell_type1")
        self.assertTrue(np.all(prediction.values == prediction2.values))


if __name__ == "__main__":
    # BLASTTest.setUpClass()
    # test = BLASTTest()
    # test.setUp()
    # test.test_npdv1_dist_blast()
    # test.tearDown()
    unittest.main()



#!/usr/bin/env python

import sys
import os
import shutil
import unittest
import numpy as np
import matplotlib
matplotlib.use("agg")

if os.environ.get("TEST_MODE", "INSTALL") == "DEV":
    sys.path.insert(0, "..")
import Cell_BLAST as cb
cb.config.RANDOM_SEED = 0

os.environ["TF_CPP_MIN_LOG_LEVEL"] = "2"


class DirectiTest(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.data = cb.data.ExprDataSet.read_dataset(
            "pollen.h5"
        ).normalize()

    def tearDown(self):
        if os.path.exists("./test_directi"):
            shutil.rmtree("./test_directi")

    def test_gau(self):
        model = cb.directi.fit_DIRECTi(
            self.data, latent_dim=10, epoch=3, path="./test_directi"
        )
        self.data.latent = model.inference(self.data)
        self.assertFalse(np.any(np.isnan(self.data.latent)))
        with self.assertRaises(Exception):
            model.clustering(self.data)
        _ = self.data.visualize_latent("cell_type1", method="tSNE", sort=True, dr_kws=dict(n_iter=250))
        _ = self.data.visualize_latent("cell_type1", method="tSNE", reuse=False, sort=True, dr_kws=dict(n_iter=250))
        _ = self.data.visualize_latent("cell_type1", method="UMAP", random_seed=123, dr_kws=dict(n_epochs=20))
        _ = self.data.visualize_latent("cell_type1", method="UMAP", random_seed=123, dr_kws=dict(n_epochs=20))
        _ = self.data.visualize_latent("cell_type1", method=None)
        with self.assertRaises(ValueError):
            _ = self.data.visualize_latent("cell_type1", method="NA")
        model.save()
        model.close()
        del model
        model = cb.directi.DIRECTi.load("./test_directi")
        latent2 = model.inference(self.data)
        self.assertTrue(np.all(self.data.latent == latent2))

    def test_catgau(self):
        with self.assertRaises(ValueError):
            _ = self.data.visualize_latent("cell_type1", method="tSNE", dr_kws=dict(n_iter=250))
        model = cb.directi.DIRECTi(
            genes=self.data.uns["scmap_genes"],
            latent_module=cb.latent.CatGau(
                latent_dim=10, cat_dim=20, multiclass_adversarial=True,
                cat_merge=True, min_silhouette=0.2, patience=2
            ),
            rmbatch_modules=[],
            prob_module=cb.prob.MSE(),
            path="./test_directi"
        ).compile("RMSPropOptimizer", 1e-3)
        model.fit(
            cb.utils.DataDict(
                exprs=self.data[:, self.data.uns["scmap_genes"]].exprs,
                library_size=np.array(self.data.exprs.sum(axis=1)).reshape((-1, 1))
            ), epoch=300, patience=20
        )
        _ = model.clustering(self.data)
        self.data.latent = model.inference(self.data)
        self.assertFalse(np.any(np.isnan(self.data.latent)))
        model.save()
        model.close()
        del model
        model = cb.directi.DIRECTi.load("./test_directi")
        latent2 = model.inference(self.data)
        self.assertTrue(np.all(self.data.latent == latent2))
        random_state = np.random.RandomState(0)
        model.gene_grad(
            self.data, random_state.randn(self.data.shape[0], 10)
        )

    def test_semisupervised_catgau(self):
        _ = self.data.annotation_confidence(
            "cell_type1", used_vars="scmap_genes",
            return_group_percentile=False
        )
        self.data.obs.loc[
            self.data.annotation_confidence(
                "cell_type1", return_group_percentile=True
            )[1] <= 0.5, "cell_type1"
        ] = ""
        model = cb.directi.fit_DIRECTi(
            self.data, genes=self.data.uns["scmap_genes"],
            latent_dim=10, cat_dim=20, prob_module="ZINB",
            supervision="cell_type1", epoch=3, path="./test_directi"
        )
        self.data.latent = model.inference(self.data)
        self.assertFalse(np.any(np.isnan(self.data.latent)))
        model.save()
        model.close()
        del model
        model = cb.directi.DIRECTi.load("./test_directi")
        latent2 = model.inference(self.data)
        self.assertTrue(np.all(self.data.latent == latent2))

    def test_rmbatch(self):
        model = cb.directi.fit_DIRECTi(
            self.data, genes=self.data.uns["scmap_genes"],
            latent_dim=10, batch_effect="cell_type1",  # Just for test
            epoch=3, path="./test_directi"
        )
        self.data.latent = model.inference(self.data)
        self.assertFalse(np.any(np.isnan(self.data.latent)))
        model.save()
        model.close()
        del model
        model = cb.directi.DIRECTi.load("./test_directi")
        latent2 = model.inference(self.data)
        self.assertTrue(np.all(self.data.latent == latent2))

        model = cb.directi.fit_DIRECTi(
            self.data, genes=self.data.uns["scmap_genes"],
            prob_module="NB", latent_dim=10, depth=0,
            batch_effect="cell_type1",  # Just for test
            rmbatch_module="RMBatch",
            epoch=3, path="./test_directi"
        )
        self.data.latent = model.inference(self.data)
        self.assertFalse(np.any(np.isnan(self.data.latent)))
        model.save()
        model.close()
        del model
        model = cb.directi.DIRECTi.load("./test_directi")
        latent2 = model.inference(self.data)
        self.assertTrue(np.all(self.data.latent == latent2))

        model = cb.directi.fit_DIRECTi(
            self.data, genes=self.data.uns["scmap_genes"],
            prob_module="ZINB",
            latent_dim=10, batch_effect="cell_type1",  # Just for test
            rmbatch_module="MNN",
            epoch=3, path="./test_directi"
        )
        self.data.latent = model.inference(self.data)
        self.assertFalse(np.any(np.isnan(self.data.latent)))
        model.save()
        model.close()
        del model
        model = cb.directi.DIRECTi.load("./test_directi")
        latent2 = model.inference(self.data)
        self.assertTrue(np.all(self.data.latent == latent2))

        model = cb.directi.fit_DIRECTi(
            self.data, genes=self.data.uns["scmap_genes"],
            prob_module="ZILN",
            latent_dim=10, batch_effect="cell_type1",  # Just for test
            rmbatch_module="MNNAdversarial",
            epoch=3, path="./test_directi"
        )
        self.data.latent = model.inference(self.data)
        self.assertFalse(np.any(np.isnan(self.data.latent)))
        model.save()
        model.close()
        del model
        model = cb.directi.DIRECTi.load("./test_directi")
        latent2 = model.inference(self.data)
        self.assertTrue(np.all(self.data.latent == latent2))

        model = cb.directi.fit_DIRECTi(
            self.data, genes=self.data.uns["scmap_genes"],
            prob_module="LN",
            latent_dim=10, batch_effect="cell_type1",  # Just for test
            rmbatch_module="AdaptiveMNNAdversarial",
            epoch=3, path="./test_directi"
        )
        self.data.latent = model.inference(self.data)
        self.assertFalse(np.any(np.isnan(self.data.latent)))
        model.save()
        model.close()
        del model
        model = cb.directi.DIRECTi.load("./test_directi")
        latent2 = model.inference(self.data)
        self.assertTrue(np.all(self.data.latent == latent2))


if __name__ == "__main__":
    # DirectiTest.setUpClass()
    # test = DirectiTest()
    # test.test_catgau()
    # test.tearDown()
    unittest.main()



r"""
Model initialization, training and saving/loading framework used by DIRECTi
"""


import json
import os
import abc
import time
import traceback
import typing
import tempfile
from builtins import input

import numpy as np
import tensorflow as tf

from . import utils


class Model(object):
    r"""
    Abstract model class, providing a framework for model initialization,
    training, saving and loading.
    """
    def __init__(
            self, random_seed: typing.Optional[int] = None,
            path: typing.Optional[str] = None, **kwargs
    ) -> None:
        self.random_state = np.random.RandomState(random_seed)
        self.graph = tf.Graph()
        self.vars_to_save = []
        with self.graph.as_default():  # pylint: disable=not-context-manager
            if random_seed is not None:
                tf.set_random_seed(random_seed)
            self._init_graph(**kwargs)
            self._init_session()
        if path is None:
            self.path = tempfile.mkdtemp()
        else:
            os.makedirs(path, exist_ok=True)
            self.path = path
        utils.logger.info("Using model path: %s", self.path)

    @utils.with_self_graph
    def compile(
            self, optimizer: str, lr: float, initialize_weights: bool = True
    ) -> "Model":
        r"""
        Compile the model and get ready for fitting.

        Parameters
        ----------
        optimizer
            Name of the optimizer to use.
        lr
            Learning rate.
        initialize_weights
            Whether to initialize model weights.

        Returns
        -------
        model
            A compiled model.
        """
        self._compile(optimizer, lr)
        self._summarize()
        if initialize_weights:
            self.sess.run(tf.global_variables_initializer())
        return self

    # Called with self.graph.as_default
    def _init_session(self) -> None:
        config = tf.ConfigProto()
        config.gpu_options.allow_growth = True  # pylint: disable=no-member
        self.sess = tf.Session(config=config)

    # Called with self.graph.as_default
    def _init_graph(self) -> None:
        with tf.variable_scope("epoch", reuse=tf.AUTO_REUSE):
            self.epoch = tf.get_variable(
                "epoch", shape=(), dtype=tf.int32, trainable=False)

    # Called with self.graph.as_default
    @abc.abstractmethod
    def _compile(self, optimizer, lr) -> None:  # pragma: no cover
        raise NotImplementedError

    # Called with self.graph.as_default
    def _summarize(self) -> None:
        self.summarizer = tf.summary.FileWriter(
            os.path.join(self.path, "summary"),
            graph=self.graph, flush_secs=10
        )

    def close(self) -> None:
        r"""
        Clean up and close the model.
        """
        self.sess.close()
        # tf.reset_default_graph()

    @utils.with_self_graph
    def fit(
            self, data_dict: utils.DataDict,
            val_split: float = 0.1, epoch: int = 100,
            patience: int = np.inf, tolerance: float = 0.0,
            on_epoch_end: typing.Optional[typing.List[typing.Callable[
                ["Model", utils.DataDict, utils.DataDict, float], bool
            ]]] = None,
            progress_bar: bool = True, **kwargs
    ) -> "Model":
        r"""
        This function wraps an epoch-by-epoch update function into
        complete training process that supports data splitting, shuffling
        and early stop.
        """
        if on_epoch_end is None:
            on_epoch_end = []
        data_dict = utils.DataDict(data_dict)

        # Leave out validation set
        data_dict = data_dict.shuffle(self.random_state)
        data_size = data_dict.size
        train_data_dict = data_dict[int(val_split * data_size):]
        val_data_dict = data_dict[:int(val_split * data_size)]

        # Fit preparation
        loss_record = np.inf
        patience_countdown = patience
        saver = tf.train.Saver(max_to_keep=1)
        ckpt_file = os.path.join(self.path, "checkpoint")

        # Fit loop
        for epoch_idx in range(epoch):
            self.epoch_report = ""
            try:
                self.epoch_report += f"[{self.__class__.__name__} epoch {epoch_idx}] "
                self.sess.run(self.epoch.assign(epoch_idx))

                try:
                    t_start = time.time()
                    self._fit_epoch(
                        train_data_dict.shuffle(self.random_state),
                        progress_bar=progress_bar, **kwargs
                    )
                    loss = self._val_epoch(
                        val_data_dict,
                        progress_bar=progress_bar, **kwargs
                    )
                    self.epoch_report += f"time elapsed={time.time() - t_start:.1f}s"
                except Exception:  # pragma: no cover
                    print("\n==== Oops! Model has crashed... ====\n")
                    traceback.print_exc()
                    print("\n====================================\n")
                    break

                all_converged = True
                for fn in on_epoch_end:
                    all_converged = fn(
                        self, train_data_dict, val_data_dict, loss
                    ) and all_converged

                # Early stop
                if all_converged:
                    if np.isfinite(patience) and loss < loss_record + tolerance:
                        self.epoch_report += " Best save..."
                        latest_checkpoint = saver.save(
                            self.sess, ckpt_file, global_step=epoch_idx,
                            write_meta_graph=False
                        )
                        patience_countdown = patience
                        if loss < loss_record:
                            loss_record = loss
                    else:
                        patience_countdown -= 1
                    print(self.epoch_report)
                    if patience_countdown == 0:
                        break
                else:
                    loss_record = np.inf  # In case all_converged is exited
                    patience_countdown = patience
                    if epoch_idx % 10 == 0:  # Save regularly
                        self.epoch_report += " Regular save..."
                        latest_checkpoint = saver.save(
                            self.sess, ckpt_file, global_step=epoch_idx,
                            write_meta_graph=False
                        )
                    print(self.epoch_report)

            except KeyboardInterrupt:  # pragma: no cover
                print("\n\n==== Caught keyboard interruption! ====\n")
                success_flag = False
                break_flag = False
                while not success_flag:
                    choice = input("Stop model training? (y/n) ")
                    if choice == "y":
                        break_flag = True
                        break
                    elif choice == "n":
                        success_flag = True
                if break_flag:
                    break

        # Fit finish
        if "latest_checkpoint" in locals():
            print("Restoring best model...")
            saver.restore(self.sess, latest_checkpoint)
        return self

    # Called with self.graph.as_default
    @abc.abstractmethod
    def _fit_epoch(
            self, data_dict: utils.DataDict, **kwargs
    ) -> None:  # pragma: no cover
        raise NotImplementedError

    # Called with self.graph.as_default
    @abc.abstractmethod
    def _val_epoch(
            self, data_dict: utils.DataDict, **kwargs
    ) -> float:  # pragma: no cover
        raise NotImplementedError

    def _get_config(self) -> typing.Mapping:
        return {}

    def _save_config(self, file: str) -> None:
        with open(file, "w") as f:
            json.dump(self._get_config(), f, indent=4)

    @classmethod
    def _load_config(cls, file: str, **kwargs) -> "Model":
        with open(file, "r") as f:
            config = json.load(f)
        return cls(path=os.path.dirname(file), **config, **kwargs)

    @utils.with_self_graph
    def _save_weights(self, path: str) -> None:
        if self.vars_to_save:
            if not os.path.exists(path):
                os.makedirs(path)
            tf.train.Saver(var_list=self.vars_to_save, max_to_keep=1).save(
                self.sess, os.path.join(path, "save.ckpt"),
                write_meta_graph=False
            )

    @utils.with_self_graph
    def _load_weights(self, path: str) -> None:
        failed_vars = []
        for var_to_save in self.vars_to_save:
            try:
                tf.train.Saver(
                    var_list=[var_to_save], max_to_keep=1
                ).restore(
                    self.sess, os.path.join(path, "save.ckpt")
                )
            except Exception:
                failed_vars.append(var_to_save)
        if failed_vars:
            utils.logger.info("%d variables failed to load.", len(failed_vars))
            utils.logger.debug(str(failed_vars))

    def save(
            self, path: typing.Optional[str] = None,
            config: str = "config.json", weights: str = "weights"
    ) -> None:
        r"""
        Save model configuration and weights.

        Parameters
        ----------
        path
            Path to save the model. If not specified, :attr:`Model.path`
            will be used.
        config
            File to store model configuration (in json format, under ``path``).
        weights
            Directory to store model weights (under ``path``).
        """
        if path is None:
            path = self.path
        elif not os.path.exists(path):
            os.makedirs(path)
        self._save_config(os.path.join(path, config))
        self._save_weights(os.path.join(path, weights))

    @classmethod
    def load(
            cls, path: str,
            config: str = "config.json", weights: str = "weights", **kwargs
    ) -> "Model":
        r"""
        Load model configuration and weights.

        Parameters
        ----------
        path
            Model path.
        config
            File that stores model configuration (in json format, under ``path``).
        weights
            Directory that stores model weights (under ``path``).
        kwargs
            Additional keyword arguments to be passed to the class constructor.

        Returns
        -------
        loaded_model
            Loaded model.
        """
        model = cls._load_config(os.path.join(path, config), **kwargs)
        model._load_weights(os.path.join(path, weights))
        return model
